from pathlib import Path
from typing import Any, Dict, Set

from transformers import CONFIG_MAPPING

from .class_factory import LightningIRConfigClassFactory
from .external_model_hub import CHECKPOINT_MAPPING


class LightningIRConfig:
    """The configuration class to instantiate a LightningIR model. Acts as a mixin for the
    transformers.PretrainedConfig_ class.

    .. _transformers.PretrainedConfig: \
https://huggingface.co/transformers/main_classes/configuration.html#transformers.PretrainedConfig
    """

    model_type = "lightning-ir"
    """Model type for the configuration."""
    backbone_model_type: str | None = None
    """Backbone model type for the configuration. Set by :func:`LightningIRModelClassFactory`."""

    TOKENIZER_ARGS: Set[str] = {"query_length", "doc_length"}
    """Arguments for the tokenizer."""
    ADDED_ARGS: Set[str] = TOKENIZER_ARGS
    """Arguments added to the configuration."""

    def __init__(self, *args, query_length: int = 32, doc_length: int = 512, **kwargs):
        """Initializes the configuration.

        :param query_length: Maximum query length, defaults to 32
        :type query_length: int, optional
        :param doc_length: Maximum document length, defaults to 512
        :type doc_length: int, optional
        """
        super().__init__(*args, **kwargs)
        self.query_length = query_length
        self.doc_length = doc_length

    def to_added_args_dict(self) -> Dict[str, Any]:
        """Outputs a dictionary of the added arguments.

        :return: Added arguments
        :rtype: Dict[str, Any]
        """
        return {arg: getattr(self, arg) for arg in self.ADDED_ARGS if hasattr(self, arg)}

    def to_tokenizer_dict(self) -> Dict[str, Any]:
        """Outputs a dictionary of the tokenizer arguments.

        :return: Tokenizer arguments
        :rtype: Dict[str, Any]
        """
        return {arg: getattr(self, arg) for arg in self.TOKENIZER_ARGS}

    def to_dict(self) -> Dict[str, Any]:
        """Overrides the transformers.PretrainedConfig.to_dict_ method to include the added arguments and the backbone
        model type.

        .. _transformers.PretrainedConfig.to_dict: \
https://huggingface.co/docs/transformers/main_classes/configuration.html#transformers.PretrainedConfig.to_dict

        :return: Configuration dictionary
        :rtype: Dict[str, Any]
        """
        if hasattr(super(), "to_dict"):
            output = getattr(super(), "to_dict")()
        else:
            output = self.to_added_args_dict()
        if self.backbone_model_type is not None:
            output["backbone_model_type"] = self.backbone_model_type
        return output

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str | Path, *args, **kwargs) -> "LightningIRConfig":
        if cls is LightningIRConfig or all(issubclass(base, LightningIRConfig) for base in cls.__bases__):
            config = None
            if pretrained_model_name_or_path in CHECKPOINT_MAPPING:
                config = CHECKPOINT_MAPPING[pretrained_model_name_or_path]
                config_class = config.__class__
            elif cls is not LightningIRConfig:
                config_class = cls
            else:
                config_class = LightningIRConfigClassFactory.get_lightning_ir_config(pretrained_model_name_or_path)
                if config_class is None:
                    raise ValueError("Pass a config to `from_pretrained`.")
            BackboneConfig = LightningIRConfigClassFactory.get_backbone_config(pretrained_model_name_or_path)
            cls = LightningIRConfigClassFactory(config_class).from_backbone_class(BackboneConfig)
            if config is not None and all(issubclass(base, LightningIRConfig) for base in config.__class__.__bases__):
                derived_config = cls.from_pretrained(pretrained_model_name_or_path, config=config)
                derived_config.update(config.to_dict())
            return cls.from_pretrained(pretrained_model_name_or_path, *args, **kwargs)
        return super(LightningIRConfig, cls).from_pretrained(pretrained_model_name_or_path, *args, **kwargs)

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any], *args, **kwargs) -> "LightningIRConfig":
        """Loads the configuration from a dictionary. Wraps the transformers.PretrainedConfig.from_dict_ method to
        return a derived LightningIRConfig class. See :class:`.LightningIRConfigClassFactory` for more details.

        .. _transformers.PretrainedConfig.from_dict: \
https://huggingface.co/docs/transformers/main_classes/configuration.html#transformers.PretrainedConfig.from_dict

        :param config_dict: Configuration dictionary
        :type config_dict: Dict[str, Any]
        :raises ValueError: If the model type does not match the configuration model type
        :return: Derived LightningIRConfig class
        :rtype: LightningIRConfig
        """
        if all(issubclass(base, LightningIRConfig) for base in cls.__bases__) or cls is LightningIRConfig:
            if "backbone_model_type" in config_dict:
                backbone_model_type = config_dict["backbone_model_type"]
                model_type = config_dict["model_type"]
                if cls is not LightningIRConfig and model_type != cls.model_type:
                    raise ValueError(
                        f"Model type {model_type} does not match configuration model type {cls.model_type}"
                    )
            else:
                backbone_model_type = config_dict["model_type"]
                model_type = cls.model_type
            MixinConfig = CONFIG_MAPPING[model_type]
            BackboneConfig = CONFIG_MAPPING[backbone_model_type]
            cls = LightningIRConfigClassFactory(MixinConfig).from_backbone_class(BackboneConfig)
            return cls.from_dict(config_dict, *args, **kwargs)
        return super(LightningIRConfig, cls).from_dict(config_dict, *args, **kwargs)



from .config import T5CrossEncoderConfig
from .model import T5CrossEncoderModel
from .tokenizer import T5CrossEncoderTokenizer

__all__ = ["T5CrossEncoderConfig", "T5CrossEncoderModel", "T5CrossEncoderTokenizer"]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Literal, Tuple

import torch

from ..bi_encoder.model import BiEncoderEmbedding
from .searcher import SearchConfig, Searcher

if TYPE_CHECKING:
    from ..bi_encoder import BiEncoderModule


class FaissSearcher(Searcher):
    def __init__(
        self, index_dir: Path | str, search_config: FaissSearchConfig, module: BiEncoderModule, use_gpu: bool = False
    ) -> None:
        import faiss

        self.search_config: FaissSearchConfig
        self.index = faiss.read_index(str(Path(index_dir) / "index.faiss"))
        ivf_index = None
        try:
            ivf_index = faiss.extract_index_ivf(self.index)
        except RuntimeError:
            pass
        if ivf_index is not None:
            ivf_index.nprobe = search_config.n_probe
            quantizer = getattr(ivf_index, "quantizer", None)
            if quantizer is not None:
                downcasted_quantizer = faiss.downcast_index(quantizer)
                hnsw = getattr(downcasted_quantizer, "hnsw", None)
                if hnsw is not None:
                    hnsw.efSearch = search_config.ef_search
        super().__init__(index_dir, search_config, module, use_gpu)

    @property
    def num_embeddings(self) -> int:
        return self.index.ntotal

    @property
    def doc_is_single_vector(self) -> bool:
        return self.num_docs == self.num_embeddings

    def _search(self, query_embeddings: BiEncoderEmbedding) -> Tuple[torch.Tensor, torch.Tensor, List[int]]:
        query_embeddings = query_embeddings.to(self.device)
        candidate_scores, candidate_doc_idcs = self.candidate_retrieval(query_embeddings)
        query_lengths = query_embeddings.scoring_mask.sum(-1)
        if self.search_config.imputation_strategy == "gather":
            doc_embeddings, doc_idcs, num_docs = self.gather_imputation(candidate_doc_idcs, query_lengths)
            doc_scores = self.module.model.score(query_embeddings, doc_embeddings, num_docs)
        else:
            doc_scores, doc_idcs, num_docs = self.intra_ranking_imputation(
                candidate_scores, candidate_doc_idcs, query_lengths
            )
        return doc_scores, doc_idcs, num_docs

    def candidate_retrieval(self, query_embeddings: BiEncoderEmbedding) -> Tuple[torch.Tensor, torch.Tensor]:
        embeddings = query_embeddings.embeddings[query_embeddings.scoring_mask]
        candidate_scores, candidate_idcs = self.index.search(embeddings.float().cpu(), self.search_config.candidate_k)
        candidate_scores = torch.from_numpy(candidate_scores)
        candidate_idcs = torch.from_numpy(candidate_idcs)
        if self.doc_is_single_vector:
            candidate_doc_idcs = candidate_idcs.to(self.cumulative_doc_lengths.device)
        else:
            candidate_doc_idcs = torch.searchsorted(
                self.cumulative_doc_lengths,
                candidate_idcs.to(self.cumulative_doc_lengths.device),
                side="right",
            )
        return candidate_scores, candidate_doc_idcs

    def gather_imputation(
        self, candidate_doc_idcs: torch.Tensor, query_lengths: torch.Tensor
    ) -> Tuple[BiEncoderEmbedding, torch.Tensor, List[int]]:
        # unique doc_idcs per query
        doc_idcs_per_query = [
            list(sorted(set(idcs.reshape(-1).tolist())))
            for idcs in torch.split(candidate_doc_idcs, query_lengths.tolist())
        ]
        num_docs = [len(idcs) for idcs in doc_idcs_per_query]
        doc_idcs = torch.tensor(sum(doc_idcs_per_query, [])).to(candidate_doc_idcs)
        unique_doc_idcs, inverse_idcs = torch.unique(doc_idcs, return_inverse=True)

        # gather all vectors for unique doc_idcs
        doc_lengths = self.doc_lengths[unique_doc_idcs]
        start_doc_idcs = self.cumulative_doc_lengths[unique_doc_idcs - 1]
        start_doc_idcs[unique_doc_idcs == 0] = 0
        all_doc_idcs = torch.cat(
            [
                torch.arange(start.item(), start.item() + length.item())
                for start, length in zip(start_doc_idcs.cpu(), doc_lengths.cpu())
            ]
        )
        all_doc_embeddings = torch.from_numpy(self.index.reconstruct_batch(all_doc_idcs))
        unique_embeddings = torch.nn.utils.rnn.pad_sequence(
            [embeddings for embeddings in torch.split(all_doc_embeddings, doc_lengths.tolist())],
            batch_first=True,
        ).to(inverse_idcs.device)
        embeddings = unique_embeddings[inverse_idcs]

        # mask out padding
        doc_lengths = doc_lengths[inverse_idcs]
        scoring_mask = torch.arange(embeddings.shape[1], device=embeddings.device) < doc_lengths[:, None]
        doc_embeddings = BiEncoderEmbedding(embeddings=embeddings, scoring_mask=scoring_mask)
        return doc_embeddings, doc_idcs, num_docs

    def intra_ranking_imputation(
        self,
        candidate_scores: torch.Tensor,
        candidate_doc_idcs: torch.Tensor,
        query_lengths: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, List[int]]:
        max_query_length = int(query_lengths.max().item())
        is_query_single_vector = max_query_length == 1

        if self.doc_is_single_vector:
            scores = candidate_scores.view(-1)
            doc_idcs = candidate_doc_idcs.view(-1)
            num_docs = torch.full((candidate_scores.shape[0],), candidate_scores.shape[1])
        else:
            # grab unique doc ids per query candidate
            query_idcs = torch.arange(query_lengths.shape[0], device=query_lengths.device).repeat_interleave(
                query_lengths
            )
            query_candidate_idcs = torch.cat(
                [torch.arange(length.item(), device=query_lengths.device) for length in query_lengths]
            )
            paired_idcs = torch.stack(
                [
                    query_idcs.repeat_interleave(candidate_scores.shape[1]),
                    query_candidate_idcs.repeat_interleave(candidate_scores.shape[1]),
                    candidate_doc_idcs.view(-1),
                ]
            ).T
            unique_paired_idcs, inverse_idcs = torch.unique(paired_idcs[:, [0, 2]], return_inverse=True, dim=0)
            doc_idcs = unique_paired_idcs[:, 1]
            num_docs = unique_paired_idcs[:, 0].bincount()

            # accumulate max score per doc
            ranking_doc_idcs = torch.arange(doc_idcs.shape[0], device=query_lengths.device)[inverse_idcs]
            idcs = ranking_doc_idcs * max_query_length + paired_idcs[:, 1]
            shape = torch.Size((doc_idcs.shape[0], max_query_length))
            scores = torch.scatter_reduce(
                torch.full((shape.numel(),), float("inf"), device=query_lengths.device),
                0,
                idcs,
                candidate_scores.view(-1).to(query_lengths.device),
                "max",
                include_self=False,
            ).view(shape)

        if is_query_single_vector:
            scores = scores.squeeze(-1)
        else:
            # impute missing values
            if self.search_config.imputation_strategy == "min":
                impute_values = (
                    scores.masked_fill(scores == torch.finfo(scores.dtype).min, float("inf"))
                    .min(0, keepdim=True)
                    .values.expand_as(scores)
                )
            elif self.search_config.imputation_strategy is None:
                impute_values = torch.zeros_like(scores)
            else:
                raise ValueError("Invalid imputation strategy: " f"{self.search_config.imputation_strategy}")
            is_inf = torch.isinf(scores)
            scores[is_inf] = impute_values[is_inf]

            # aggregate score per query vector
            mask = (
                torch.arange(max_query_length, device=query_lengths.device) < query_lengths[:, None]
            ).repeat_interleave(num_docs, dim=0)
            scores = self.module.scoring_function.aggregate(
                scores, mask, self.module.config.query_aggregation_function, dim=1
            ).squeeze(-1)
        return scores, doc_idcs, num_docs.tolist()


class FaissSearchConfig(SearchConfig):
    search_class = FaissSearcher

    def __init__(
        self,
        k: int = 10,
        candidate_k: int = 100,
        imputation_strategy: Literal["min", "gather"] | None = None,
        n_probe: int = 1,
        ef_search: int = 16,
    ) -> None:
        super().__init__(k)
        self.candidate_k = candidate_k
        self.imputation_strategy = imputation_strategy
        self.n_probe = n_probe
        self.ef_search = ef_search



import warnings
from typing import Dict, Sequence

from tokenizers.processors import TemplateProcessing
from transformers import BatchEncoding, BertTokenizer, BertTokenizerFast

from ..base import LightningIRTokenizer
from .config import BiEncoderConfig


class BiEncoderTokenizer(LightningIRTokenizer):

    config_class = BiEncoderConfig

    def __init__(
        self,
        *args,
        query_token: str = "[QUE]",
        doc_token: str = "[DOC]",
        query_expansion: bool = False,
        query_length: int = 32,
        attend_to_query_expanded_tokens: bool = False,
        doc_expansion: bool = False,
        doc_length: int = 512,
        attend_to_doc_expanded_tokens: bool = False,
        add_marker_tokens: bool = True,
        **kwargs,
    ):
        super().__init__(
            *args,
            query_token=query_token,
            doc_token=doc_token,
            query_expansion=query_expansion,
            query_length=query_length,
            attend_to_query_expanded_tokens=attend_to_query_expanded_tokens,
            doc_expansion=doc_expansion,
            doc_length=doc_length,
            attend_to_doc_expanded_tokens=attend_to_doc_expanded_tokens,
            add_marker_tokens=add_marker_tokens,
            **kwargs,
        )
        self.query_expansion = query_expansion
        self.query_length = query_length
        self.attend_to_query_expanded_tokens = attend_to_query_expanded_tokens
        self.doc_expansion = doc_expansion
        self.doc_length = doc_length
        self.attend_to_doc_expanded_tokens = attend_to_doc_expanded_tokens
        self.add_marker_tokens = add_marker_tokens

        self._query_token = query_token
        self._doc_token = doc_token

        self.query_post_processor: TemplateProcessing | None = None
        self.doc_post_processor: TemplateProcessing | None = None
        if add_marker_tokens:
            # TODO support other tokenizers
            if not isinstance(self, (BertTokenizer, BertTokenizerFast)):
                raise ValueError("Adding marker tokens is only supported for BertTokenizer.")
            self.add_tokens([query_token, doc_token], special_tokens=True)
            self.query_post_processor = TemplateProcessing(
                single=f"[CLS] {self.query_token} $0 [SEP]",
                pair=f"[CLS] {self.query_token} $A [SEP] {self.doc_token} $B:1 [SEP]:1",
                special_tokens=[
                    ("[CLS]", self.cls_token_id),
                    ("[SEP]", self.sep_token_id),
                    (self.query_token, self.query_token_id),
                    (self.doc_token, self.doc_token_id),
                ],
            )
            self.doc_post_processor = TemplateProcessing(
                single=f"[CLS] {self.doc_token} $0 [SEP]",
                pair=f"[CLS] {self.query_token} $A [SEP] {self.doc_token} $B:1 [SEP]:1",
                special_tokens=[
                    ("[CLS]", self.cls_token_id),
                    ("[SEP]", self.sep_token_id),
                    (self.query_token, self.query_token_id),
                    (self.doc_token, self.doc_token_id),
                ],
            )

    @property
    def query_token(self) -> str:
        return self._query_token

    @property
    def doc_token(self) -> str:
        return self._doc_token

    @property
    def query_token_id(self) -> int | None:
        if self.query_token in self.added_tokens_encoder:
            return self.added_tokens_encoder[self.query_token]
        return None

    @property
    def doc_token_id(self) -> int | None:
        if self.doc_token in self.added_tokens_encoder:
            return self.added_tokens_encoder[self.doc_token]
        return None

    def __call__(self, *args, warn: bool = True, **kwargs) -> BatchEncoding:
        if warn:
            warnings.warn(
                "BiEncoderTokenizer is being directly called. Use tokenize_query and "
                "tokenize_doc to make sure marker_tokens and query/doc expansion is "
                "applied."
            )
        return super().__call__(*args, **kwargs)

    def _encode(
        self,
        text: str | Sequence[str],
        *args,
        post_processor: TemplateProcessing | None = None,
        **kwargs,
    ) -> BatchEncoding:
        orig_post_processor = self._tokenizer.post_processor
        if post_processor is not None:
            self._tokenizer.post_processor = post_processor
        if kwargs.get("return_tensors", None) is not None:
            kwargs["pad_to_multiple_of"] = 8
        encoding = self(text, *args, warn=False, **kwargs)
        self._tokenizer.post_processor = orig_post_processor
        return encoding

    def _expand(self, encoding: BatchEncoding, attend_to_expanded_tokens: bool) -> BatchEncoding:
        input_ids = encoding["input_ids"]
        input_ids[input_ids == self.pad_token_id] = self.mask_token_id
        encoding["input_ids"] = input_ids
        if attend_to_expanded_tokens:
            encoding["attention_mask"].fill_(1)
        return encoding

    def tokenize_query(self, queries: Sequence[str] | str, *args, **kwargs) -> BatchEncoding:
        kwargs["max_length"] = self.query_length
        if self.query_expansion:
            kwargs["padding"] = "max_length"
        else:
            kwargs["truncation"] = True
        encoding = self._encode(queries, *args, post_processor=self.query_post_processor, **kwargs)
        if self.query_expansion:
            self._expand(encoding, self.attend_to_query_expanded_tokens)
        return encoding

    def tokenize_doc(self, docs: Sequence[str] | str, *args, **kwargs) -> BatchEncoding:
        kwargs["max_length"] = self.doc_length
        if self.doc_expansion:
            kwargs["padding"] = "max_length"
        else:
            kwargs["truncation"] = True
        encoding = self._encode(docs, *args, post_processor=self.doc_post_processor, **kwargs)
        if self.doc_expansion:
            self._expand(encoding, self.attend_to_doc_expanded_tokens)
        return encoding

    def tokenize(
        self,
        queries: str | Sequence[str] | None = None,
        docs: str | Sequence[str] | None = None,
        **kwargs,
    ) -> Dict[str, BatchEncoding]:
        encodings = {}
        kwargs.pop("num_docs", None)
        if queries is not None:
            encodings["query_encoding"] = self.tokenize_query(queries, **kwargs)
        if docs is not None:
            encodings["doc_encoding"] = self.tokenize_doc(docs, **kwargs)
        return encodings



from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Sequence

import torch
from lightning import LightningDataModule
from torch.utils.data import DataLoader, IterableDataset

from ..base.config import LightningIRConfig
from ..base.tokenizer import LightningIRTokenizer
from .data import IndexBatch, RankBatch, SearchBatch, TrainBatch
from .dataset import DocDataset, DocSample, QueryDataset, QuerySample, RankSample, RunDataset, TupleDataset

if TYPE_CHECKING:
    from ..base import LightningIRModule


class LightningIRDataModule(LightningDataModule):
    def __init__(
        self,
        model_name_or_path: str | Path | None = None,
        config: LightningIRConfig | None = None,
        module: LightningIRModule | None = None,
        num_workers: int = 0,
        train_batch_size: int | None = None,
        shuffle_train: bool = True,
        inference_batch_size: int | None = None,
        train_dataset: RunDataset | TupleDataset | None = None,
        inference_datasets: Sequence[RunDataset | TupleDataset | QueryDataset | DocDataset] | None = None,
    ) -> None:
        super().__init__()
        if config is not None:
            self.config = config
        elif module is not None:
            self.config = module.config
        elif model_name_or_path is not None:
            self.config = LightningIRConfig.from_pretrained(model_name_or_path)
        else:
            raise ValueError("Either module, config, or model_name_or_path must be provided.")

        if model_name_or_path is None:
            model_name_or_path = self.config.name_or_path
        self.tokenizer = LightningIRTokenizer.from_pretrained(model_name_or_path, config=self.config)
        self.num_workers = num_workers

        self.train_batch_size = train_batch_size
        self.shuffle_train = shuffle_train
        self.inference_batch_size = inference_batch_size
        self.train_dataset = train_dataset
        self.inference_datasets = inference_datasets

    def setup_inference(self, stage: Literal["validate", "test"]) -> None:
        if self.inference_datasets is None:
            return
        for inference_dataset in self.inference_datasets:
            if isinstance(inference_dataset, TupleDataset):
                if stage == "test":
                    raise ValueError("Prediction cannot be performed with TupleDataset.")
            elif isinstance(inference_dataset, RunDataset):
                if inference_dataset.sampling_strategy == "single_relevant":
                    raise ValueError("Inference RunDataset cannot use the single_relevant sampling strategy.")
            elif isinstance(inference_dataset, (QueryDataset, DocDataset)):
                pass
            else:
                raise ValueError(
                    "Inference Dataset must be of type RunDataset, TupleDataset, QueryDataset, or DocDataset."
                )

    def setup(self, stage: Literal["fit", "validate", "test"]) -> None:
        if stage == "fit":
            if self.train_dataset is None:
                raise ValueError("A training dataset and config must be provided.")
        if stage == "fit":
            stage = "validate"
        self.setup_inference(stage)

    def train_dataloader(self) -> DataLoader:
        if self.train_dataset is None:
            raise ValueError("No training dataset found.")
        return DataLoader(
            self.train_dataset,
            batch_size=self.train_batch_size,
            num_workers=self.num_workers,
            collate_fn=self.collate_fn,
            shuffle=(False if isinstance(self.train_dataset, IterableDataset) else self.shuffle_train),
            prefetch_factor=16 if self.num_workers > 0 else None,
        )

    def val_dataloader(self) -> List[DataLoader]:
        return self.inference_dataloader()

    def test_dataloader(self) -> List[DataLoader]:
        return self.inference_dataloader()

    def inference_dataloader(self) -> List[DataLoader]:
        inference_datasets = self.inference_datasets or []
        return [
            DataLoader(
                dataset,
                batch_size=self.inference_batch_size,
                num_workers=self.num_workers,
                collate_fn=self.collate_fn,
                prefetch_factor=16 if self.num_workers > 0 else None,
            )
            for dataset in inference_datasets
        ]

    def _aggregate_samples(self, samples: Sequence[RankSample | QuerySample | DocSample]) -> Dict[str, Any]:
        aggregated = defaultdict(list)
        field_options = {
            "query_id": {"extend": False},
            "query": {"extend": False},
            "doc_id": {"extend": False},
            "doc_ids": {"extend": False},
            "doc": {"extend": False},
            "docs": {"extend": False},
            "targets": {"extend": True},
            "qrels": {"extend": True},
        }
        for sample in samples:
            for field in sample.__dict__:
                extend = field_options[field]["extend"]
                key = field if field.endswith("s") else f"{field}s"
                value = getattr(sample, field)
                if value is None:
                    continue
                if extend:
                    aggregated[key].extend(value)
                else:
                    aggregated[key].append(value)
        return aggregated

    def _clean_sample(self, aggregated: Dict[str, Any]) -> Dict[str, Any]:
        kwargs: Dict[str, Any] = dict(aggregated)
        if "querys" in kwargs:
            kwargs["queries"] = kwargs["querys"]
            del kwargs["querys"]
        if "targets" in kwargs:
            kwargs["targets"] = torch.stack(kwargs["targets"])
        return kwargs

    def _parse_batch(
        self, sample: RankSample | QuerySample | DocSample, **kwargs
    ) -> RankBatch | TrainBatch | IndexBatch | SearchBatch:
        if isinstance(sample, RankSample):
            if "targets" in kwargs:
                return TrainBatch(**kwargs)
            else:
                return RankBatch(**kwargs)
        if isinstance(sample, QuerySample):
            return SearchBatch(**kwargs)
        if isinstance(sample, DocSample):
            return IndexBatch(**kwargs)
        raise ValueError("Invalid dataset configuration.")

    def collate_fn(
        self,
        samples: Sequence[RankSample | QuerySample | DocSample] | RankSample | QuerySample | DocSample,
    ) -> TrainBatch | RankBatch | IndexBatch | SearchBatch:
        if isinstance(samples, (RankSample, QuerySample, DocSample)):
            samples = [samples]
        aggregated = self._aggregate_samples(samples)
        kwargs = self._clean_sample(aggregated)
        return self._parse_batch(samples[0], **kwargs)



from typing import Literal

from ..col import ColConfig


class XTRConfig(ColConfig):
    model_type = "xtr"

    ADDED_ARGS = ColConfig.ADDED_ARGS.union({"token_retrieval_k", "fill_strategy", "normalization"})

    def __init__(
        self,
        token_retrieval_k: int | None = None,
        fill_strategy: Literal["zero", "min"] = "zero",
        normalization: Literal["Z"] | None = "Z",
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.token_retrieval_k = token_retrieval_k
        self.fill_strategy = fill_strategy
        self.normalization = normalization



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal, Tuple

import torch

from .searcher import SearchConfig, Searcher
from .sparse_indexer import SparseIndexConfig

if TYPE_CHECKING:
    from ..bi_encoder import BiEncoderEmbedding, BiEncoderModule


class SparseIndex:
    def __init__(self, index_dir: Path, similarity_function: Literal["dot", "cosine"], use_gpu: bool = False) -> None:
        self.index = torch.load(index_dir / "index.pt")
        self.config = SparseIndexConfig.from_pretrained(index_dir)
        if similarity_function == "dot":
            self.similarity_function = self.dot_similarity
        elif similarity_function == "cosine":
            self.similarity_function = self.cosine_similarity
        else:
            raise ValueError("Unknown similarity function")
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

    def score(self, embeddings: torch.Tensor) -> torch.Tensor:
        embeddings = embeddings.to(self.device)
        similarity = self.similarity_function(embeddings, self.index).to_dense()
        return similarity

    @property
    def num_embeddings(self) -> int:
        return self.index.shape[0]

    def cosine_similarity(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        dot_product = self.dot_similarity(x, y)
        dot_product = dot_product / (torch.norm(x, dim=-1) * torch.norm(y, dim=-1))
        return -1 * torch.cdist(x, y).squeeze(-2)

    def dot_similarity(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return torch.matmul(y, x.T).T

    def to_gpu(self) -> None:
        self.index = self.index.to(self.device)


class SparseSearcher(Searcher):
    def __init__(
        self,
        index_dir: Path,
        search_config: SparseSearchConfig,
        module: BiEncoderModule,
        use_gpu: bool = True,
    ) -> None:
        self.search_config: SparseSearchConfig
        self.index = SparseIndex(index_dir, module.config.similarity_function, use_gpu)
        super().__init__(index_dir, search_config, module, use_gpu)
        self.doc_token_idcs = (
            torch.arange(self.doc_lengths.shape[0]).to(self.doc_lengths).repeat_interleave(self.doc_lengths)
        )
        self.use_gpu = use_gpu
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

    @property
    def doc_is_single_vector(self) -> bool:
        return self.cumulative_doc_lengths[-1].item() == self.cumulative_doc_lengths.shape[0]

    def to_gpu(self) -> None:
        super().to_gpu()
        self.index.to_gpu()

    @property
    def num_embeddings(self) -> int:
        return self.index.num_embeddings

    def _search(self, query_embeddings: BiEncoderEmbedding) -> Tuple[torch.Tensor, None, None]:
        embeddings = query_embeddings.embeddings[query_embeddings.scoring_mask]
        query_lengths = query_embeddings.scoring_mask.sum(-1)
        scores = self.index.score(embeddings)

        # aggregate doc token scores
        if not self.doc_is_single_vector:
            scores = torch.scatter_reduce(
                torch.zeros(scores.shape[0], self.num_docs, device=scores.device),
                1,
                self.doc_token_idcs[None].expand_as(scores),
                scores,
                "amax",
            )

        # aggregate query token scores
        query_is_single_vector = (query_lengths == 1).all()
        if not query_is_single_vector:
            query_token_idcs = torch.arange(query_lengths.shape[0]).to(query_lengths).repeat_interleave(query_lengths)
            scores = torch.scatter_reduce(
                torch.zeros(query_lengths.shape[0], self.num_docs, device=scores.device),
                0,
                query_token_idcs[:, None].expand_as(scores),
                scores,
                self.module.config.query_aggregation_function,
            )
        scores = scores.reshape(-1)
        return scores, None, None


class SparseSearchConfig(SearchConfig):
    search_class = SparseSearcher



import torch
from transformers import BatchEncoding

from ...cross_encoder.model import CrossEncoderModel, CrossEncoderOutput
from .config import T5CrossEncoderConfig


class ScaleLinear(torch.nn.Linear):

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        # Rescale output before projecting on vocab
        # See https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/transformer/transformer.py#L586 # noqa
        input = input * (input.shape[-1] ** -0.5)
        return super().forward(input)


class T5CrossEncoderModel(CrossEncoderModel):
    config_class = T5CrossEncoderConfig

    _tied_weights_keys = ["encoder.embed_tokens.weight", "decoder.embed_tokens.weight", "linear.weight"]

    def __init__(self, config: T5CrossEncoderConfig, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        self.config: T5CrossEncoderConfig
        if self.config.decoder_strategy == "mono":
            self.linear = ScaleLinear(config.hidden_size, 2, bias=config.linear_bias)
        else:
            self.linear = ScaleLinear(config.hidden_size, 1, bias=config.linear_bias)

    # TODO tieing of weights does not work when setting linear to only use slice of lm head for efficiency
    # def get_output_embeddings(self):
    #     shared = self.shared
    #     if self.config.decoder_strategy == "mono":
    #         self.linear.weight.data = shared.weight.data[[1176, 6136]]
    #     elif self.config.decoder_strategy == "rank":
    #         self.linear.weight.data = shared.weight.data[[32089]]
    #     else:
    #         raise ValueError("Unknown decoder strategy")
    #     return shared

    def forward(self, encoding: BatchEncoding) -> CrossEncoderOutput:
        decoder_input_ids = torch.zeros(
            (encoding["input_ids"].shape[0], 1), device=encoding["input_ids"].device, dtype=torch.long
        )
        encoding["decoder_input_ids"] = decoder_input_ids
        output = super().forward(encoding)
        if output.scores is None:
            raise ValueError("Scores are None")
        if self.config.decoder_strategy == "mono":
            scores = output.scores.view(-1, 2)
            scores = torch.nn.functional.log_softmax(scores, dim=-1)[:, 0]
            output.scores = scores.view(-1)
        return output



from typing import List, Sequence, Tuple

import torch

from ..base.module import LightningIRModule
from ..data import RankBatch, SearchBatch, TrainBatch
from ..loss.loss import InBatchLossFunction, LossFunction
from .config import CrossEncoderConfig
from .model import CrossEncoderModel, CrossEncoderOutput
from .tokenizer import CrossEncoderTokenizer


class CrossEncoderModule(LightningIRModule):
    def __init__(
        self,
        model_name_or_path: str | None = None,
        config: CrossEncoderConfig | None = None,
        model: CrossEncoderModel | None = None,
        loss_functions: Sequence[LossFunction | Tuple[LossFunction, float]] | None = None,
        evaluation_metrics: Sequence[str] | None = None,
    ):
        super().__init__(model_name_or_path, config, model, loss_functions, evaluation_metrics)
        self.model: CrossEncoderModel
        self.config: CrossEncoderConfig
        self.tokenizer: CrossEncoderTokenizer

    def forward(self, batch: RankBatch | TrainBatch | SearchBatch) -> CrossEncoderOutput:
        if isinstance(batch, SearchBatch):
            raise NotImplementedError("Searching is not available for cross-encoders")
        queries = batch.queries
        docs = [d for docs in batch.docs for d in docs]
        num_docs = [len(docs) for docs in batch.docs]
        encoding = self.prepare_input(queries, docs, num_docs)
        output = self.model.forward(encoding["encoding"])
        return output

    def compute_losses(self, batch: TrainBatch) -> List[torch.Tensor]:
        if self.loss_functions is None:
            raise ValueError("loss_functions must be set in the module")
        output = self.forward(batch)
        scores = output.scores
        if scores is None or batch.targets is None:
            raise ValueError("scores and targets must be set in the output and batch")

        scores = scores.view(len(batch.query_ids), -1)
        targets = batch.targets.view(*scores.shape, -1)

        losses = []
        for loss_function, _ in self.loss_functions:
            if isinstance(loss_function, InBatchLossFunction):
                raise NotImplementedError("InBatchLossFunction not implemented for cross-encoders")
            losses.append(loss_function.compute_loss(scores, targets))
        return losses



from .base import CHECKPOINT_MAPPING, POST_LOAD_CALLBACKS, STATE_DICT_KEY_MAPPING, LightningIRModel
from .bi_encoder import BiEncoderConfig
from .models import ColConfig, SpladeConfig, T5CrossEncoderConfig


def _map_colbert_marker_tokens(model: LightningIRModel) -> LightningIRModel:
    config = model.config
    query_token_id = config.vocab_size
    doc_token_id = config.vocab_size + 1
    model.resize_token_embeddings(config.vocab_size + 2, 8)
    embeddings = model.embeddings.word_embeddings.weight.data
    embeddings[query_token_id] = embeddings[1]  # [unused0]
    embeddings[doc_token_id] = embeddings[2]  # [unused1]
    return model


def _map_mono_t5_weights(model: LightningIRModel) -> LightningIRModel:
    # [1176, 6136] true, false
    model.linear.weight.data = model.shared.weight.data[[1176, 6136]]
    return model


def _map_rank_t5_weights(model: LightningIRModel) -> LightningIRModel:
    # 32089 <extra_id_10>
    model.linear.weight.data = model.shared.weight.data[[32089]]
    return model


def _register_external_models():
    CHECKPOINT_MAPPING.update(
        {
            "colbert-ir/colbertv2.0": ColConfig(
                query_length=32, doc_length=184, add_marker_tokens=True, normalize=True
            ),
            "naver/splade-v3": SpladeConfig(),
            "sentence-transformers/msmarco-bert-base-dot-v5": BiEncoderConfig(projection=None, embedding_dim=768),
            "sentence-transformers/msmarco-MiniLM-L-6-v3": BiEncoderConfig(projection=None, embedding_dim=384),
            "castorini/monot5-base-msmarco-10k": T5CrossEncoderConfig(decoder_strategy="mono"),
            "castorini/monot5-base-msmarco": T5CrossEncoderConfig(decoder_strategy="mono"),
            "castorini/monot5-large-msmarco-10k": T5CrossEncoderConfig(decoder_strategy="mono"),
            "castorini/monot5-large-msmarco": T5CrossEncoderConfig(decoder_strategy="mono"),
            "castorini/monot5-3b-msmarco-10k": T5CrossEncoderConfig(decoder_strategy="mono"),
            "castorini/monot5-3b-msmarco": T5CrossEncoderConfig(decoder_strategy="mono"),
            "Soyoung97/RankT5-base": T5CrossEncoderConfig(decoder_strategy="rank"),
            "Soyoung97/RankT5-large": T5CrossEncoderConfig(decoder_strategy="rank"),
            "Soyoung97/RankT5-3b": T5CrossEncoderConfig(decoder_strategy="rank"),
        }
    )
    STATE_DICT_KEY_MAPPING.update(
        {
            "colbert-ir/colbertv2.0": [("linear.weight", "bert.projection.weight")],
            "castorini/monot5-base-msmarco-10k": [(None, "linear.weight")],
            "castorini/monot5-base-msmarco": [(None, "linear.weight")],
            "castorini/monot5-large-msmarco-10k": [(None, "linear.weight")],
            "castorini/monot5-large-msmarco": [(None, "linear.weight")],
            "castorini/monot5-3b-msmarco-10k": [(None, "linear.weight")],
            "castorini/monot5-3b-msmarco": [(None, "linear.weight")],
            "Soyoung97/RankT5-base": [(None, "linear.weight")],
            "Soyoung97/RankT5-large": [(None, "linear.weight")],
            "Soyoung97/RankT5-3b": [(None, "linear.weight")],
        }
    )
    POST_LOAD_CALLBACKS.update(
        {
            "colbert-ir/colbertv2.0": _map_colbert_marker_tokens,
            "castorini/monot5-base-msmarco-10k": _map_mono_t5_weights,
            "castorini/monot5-base-msmarco": _map_mono_t5_weights,
            "castorini/monot5-large-msmarco-10k": _map_mono_t5_weights,
            "castorini/monot5-large-msmarco": _map_mono_t5_weights,
            "castorini/monot5-3b-msmarco-10k": _map_mono_t5_weights,
            "castorini/monot5-3b-msmarco": _map_mono_t5_weights,
            "Soyoung97/RankT5-base": _map_rank_t5_weights,
            "Soyoung97/RankT5-large": _map_rank_t5_weights,
            "Soyoung97/RankT5-3b": _map_rank_t5_weights,
        }
    )



from __future__ import annotations

import array
import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

import torch

if TYPE_CHECKING:
    from ..bi_encoder import BiEncoderConfig, BiEncoderOutput
    from ..data import IndexBatch


class Indexer(ABC):
    def __init__(
        self,
        index_dir: Path,
        index_config: IndexConfig,
        bi_encoder_config: BiEncoderConfig,
        verbose: bool = False,
    ) -> None:
        self.index_dir = index_dir
        self.index_config = index_config
        self.bi_encoder_config = bi_encoder_config
        self.doc_ids = []
        self.doc_lengths = array.array("I")
        self.num_embeddings = 0
        self.num_docs = 0
        self.verbose = verbose

    @abstractmethod
    def add(self, index_batch: IndexBatch, output: BiEncoderOutput) -> None: ...

    def save(self) -> None:
        self.index_config.save(self.index_dir)
        (self.index_dir / "doc_ids.txt").write_text("\n".join(self.doc_ids))
        doc_lengths = torch.tensor(self.doc_lengths)
        torch.save(doc_lengths, self.index_dir / "doc_lengths.pt")


class IndexConfig:
    indexer_class = Indexer

    @classmethod
    def from_pretrained(cls, index_dir: Path) -> "IndexConfig":
        with open(index_dir / "config.json", "r") as f:
            data = json.load(f)
            if data["index_type"] != cls.__name__:
                raise ValueError(f"Expected index_type {cls.__name__}, got {data['index_type']}")
            data.pop("index_type", None)
            data.pop("index_dir", None)
            return cls(**data)

    def save(self, index_dir: Path) -> None:
        index_dir.mkdir(parents=True, exist_ok=True)
        with open(index_dir / "config.json", "w") as f:
            data = self.__dict__.copy()
            data["index_dir"] = str(index_dir)
            data["index_type"] = self.__class__.__name__
            json.dump(data, f)



from abc import ABC, abstractmethod
from typing import Literal, Tuple

import torch


class LossFunction(ABC):
    @abstractmethod
    def compute_loss(self, *args, **kwargs) -> torch.Tensor: ...


class ScoringLossFunction(LossFunction):
    @abstractmethod
    def compute_loss(
        self,
        scores: torch.Tensor,
        targets: torch.Tensor,
    ) -> torch.Tensor: ...

    def process_targets(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        if targets.ndim > scores.ndim:
            return targets.max(-1).values
        return targets


class EmbeddingLossFunction(LossFunction):
    @abstractmethod
    def compute_loss(
        self,
        query_embeddings: torch.Tensor,
        doc_embeddings: torch.Tensor,
    ) -> torch.Tensor: ...


class PairwiseLossFunction(ScoringLossFunction):
    def get_pairwise_idcs(self, targets: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        # positive items are items where label is greater than other label in sample
        return torch.nonzero(targets[..., None] > targets[:, None], as_tuple=True)


class ListwiseLossFunction(ScoringLossFunction):
    pass


class MarginMSE(PairwiseLossFunction):
    def __init__(self, margin: float | Literal["scores"] = 1.0):
        self.margin = margin

    def compute_loss(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        targets = self.process_targets(scores, targets)
        query_idcs, pos_idcs, neg_idcs = self.get_pairwise_idcs(targets)
        pos = scores[query_idcs, pos_idcs]
        neg = scores[query_idcs, neg_idcs]
        margin = pos - neg
        if isinstance(self.margin, float):
            target_margin = torch.tensor(self.margin, device=scores.device)
        elif self.margin == "scores":
            target_margin = targets[query_idcs, pos_idcs] - targets[query_idcs, neg_idcs]
        else:
            raise ValueError("invalid margin type")
        loss = torch.nn.functional.mse_loss(margin, target_margin.clamp(min=0))
        return loss


class ConstantMarginMSE(MarginMSE):
    def __init__(self, margin: float = 1.0):
        super().__init__(margin)


class SupervisedMarginMSE(MarginMSE):
    def __init__(self):
        super().__init__("scores")


class RankNet(PairwiseLossFunction):
    def compute_loss(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        targets = self.process_targets(scores, targets)
        query_idcs, pos_idcs, neg_idcs = self.get_pairwise_idcs(targets)
        pos = scores[query_idcs, pos_idcs]
        neg = scores[query_idcs, neg_idcs]
        margin = pos - neg
        loss = torch.nn.functional.binary_cross_entropy_with_logits(margin, torch.ones_like(margin))
        return loss


class KLDivergence(ListwiseLossFunction):
    def compute_loss(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        targets = self.process_targets(scores, targets)
        scores = torch.nn.functional.log_softmax(scores, dim=-1)
        targets = torch.nn.functional.log_softmax(targets.to(scores), dim=-1)
        loss = torch.nn.functional.kl_div(scores, targets, log_target=True, reduction="batchmean")
        return loss


class LocalizedContrastiveEstimation(ListwiseLossFunction):
    def compute_loss(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        targets = self.process_targets(scores, targets)
        targets = targets.argmax(dim=1)
        loss = torch.nn.functional.cross_entropy(scores, targets)
        return loss


class ApproxLossFunction(ListwiseLossFunction):
    def __init__(self, temperature: float = 1) -> None:
        super().__init__()
        self.temperature = temperature

    @staticmethod
    def get_approx_ranks(scores: torch.Tensor, temperature: float) -> torch.Tensor:
        score_diff = scores[:, None] - scores[..., None]
        normalized_score_diff = torch.sigmoid(score_diff / temperature)
        # set diagonal to 0
        normalized_score_diff = normalized_score_diff * (1 - torch.eye(scores.shape[1], device=scores.device))
        approx_ranks = normalized_score_diff.sum(-1) + 1
        return approx_ranks


class ApproxNDCG(ApproxLossFunction):
    def __init__(self, temperature: float = 1, scale_gains: bool = True):
        super().__init__(temperature)
        self.scale_gains = scale_gains

    @staticmethod
    def get_dcg(
        ranks: torch.Tensor,
        targets: torch.Tensor,
        k: int | None = None,
        scale_gains: bool = True,
    ) -> torch.Tensor:
        log_ranks = torch.log2(1 + ranks)
        discounts = 1 / log_ranks
        if scale_gains:
            gains = 2**targets - 1
        else:
            gains = targets
        dcgs = gains * discounts
        if k is not None:
            dcgs = dcgs.masked_fill(ranks > k, 0)
        return dcgs.sum(dim=-1)

    @staticmethod
    def get_ndcg(
        ranks: torch.Tensor,
        targets: torch.Tensor,
        k: int | None = None,
        scale_gains: bool = True,
        optimal_targets: torch.Tensor | None = None,
    ) -> torch.Tensor:
        targets = targets.clamp(min=0)
        if optimal_targets is None:
            optimal_targets = targets
        optimal_ranks = torch.argsort(torch.argsort(optimal_targets, descending=True))
        optimal_ranks = optimal_ranks + 1
        dcg = ApproxNDCG.get_dcg(ranks, targets, k, scale_gains)
        idcg = ApproxNDCG.get_dcg(optimal_ranks, optimal_targets, k, scale_gains)
        ndcg = dcg / (idcg.clamp(min=1e-12))
        return ndcg

    def compute_loss(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        targets = self.process_targets(scores, targets)
        approx_ranks = self.get_approx_ranks(scores, self.temperature)
        ndcg = self.get_ndcg(approx_ranks, targets, k=None, scale_gains=self.scale_gains)
        loss = 1 - ndcg
        return loss.mean()


class ApproxMRR(ApproxLossFunction):
    def __init__(self, temperature: float = 1):
        super().__init__(temperature)

    @staticmethod
    def get_mrr(ranks: torch.Tensor, targets: torch.Tensor, k: int | None = None) -> torch.Tensor:
        targets = targets.clamp(None, 1)
        reciprocal_ranks = 1 / ranks
        mrr = reciprocal_ranks * targets
        if k is not None:
            mrr = mrr.masked_fill(ranks > k, 0)
        mrr = mrr.max(dim=-1)[0]
        return mrr

    def compute_loss(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        targets = self.process_targets(scores, targets)
        approx_ranks = self.get_approx_ranks(scores, self.temperature)
        mrr = self.get_mrr(approx_ranks, targets, k=None)
        loss = 1 - mrr
        return loss.mean()


class ApproxRankMSE(ApproxLossFunction):
    def __init__(
        self,
        temperature: float = 1,
        discount: Literal["log2", "reciprocal"] | None = None,
    ):
        super().__init__(temperature)
        self.discount = discount

    def compute_loss(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        targets = self.process_targets(scores, targets)
        approx_ranks = self.get_approx_ranks(scores, self.temperature)
        ranks = torch.argsort(torch.argsort(targets, descending=True)) + 1
        loss = torch.nn.functional.mse_loss(approx_ranks, ranks.to(approx_ranks), reduction="none")
        if self.discount == "log2":
            weight = 1 / torch.log2(ranks + 1)
        elif self.discount == "reciprocal":
            weight = 1 / ranks
        else:
            weight = 1
        loss = loss * weight
        loss = loss.mean()
        return loss


class NeuralLossFunction(ListwiseLossFunction):
    # TODO add neural loss functions

    def __init__(self, temperature: float = 1, tol: float = 1e-5, max_iter: int = 50) -> None:
        super().__init__()
        self.temperature = temperature
        self.tol = tol
        self.max_iter = max_iter

    def neural_sort(self, scores: torch.Tensor) -> torch.Tensor:
        # https://github.com/ermongroup/neuralsort/blob/master/pytorch/neuralsort.py
        scores = scores.unsqueeze(-1)
        dim = scores.shape[1]
        one = torch.ones((dim, 1), device=scores.device)

        A_scores = torch.abs(scores - scores.permute(0, 2, 1))
        B = torch.matmul(A_scores, torch.matmul(one, torch.transpose(one, 0, 1)))
        scaling = dim + 1 - 2 * (torch.arange(dim, device=scores.device) + 1)
        C = torch.matmul(scores, scaling.to(scores).unsqueeze(0))

        P_max = (C - B).permute(0, 2, 1)
        P_hat = torch.nn.functional.softmax(P_max / self.temperature, dim=-1)

        P_hat = self.sinkhorn_scaling(P_hat)

        return P_hat

    def sinkhorn_scaling(self, mat: torch.Tensor) -> torch.Tensor:
        # https://github.com/allegro/allRank/blob/master/allrank/models/losses/loss_utils.py#L8
        idx = 0
        while True:
            if (
                torch.max(torch.abs(mat.sum(dim=2) - 1.0)) < self.tol
                and torch.max(torch.abs(mat.sum(dim=1) - 1.0)) < self.tol
            ) or idx > self.max_iter:
                break
            mat = mat / mat.sum(dim=1, keepdim=True).clamp(min=1e-12)
            mat = mat / mat.sum(dim=2, keepdim=True).clamp(min=1e-12)
            idx += 1

        return mat

    def get_sorted_targets(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        permutation_matrix = self.neural_sort(scores)
        pred_sorted_targets = torch.matmul(permutation_matrix, targets[..., None].to(permutation_matrix)).squeeze(-1)
        return pred_sorted_targets


class InBatchLossFunction(ScoringLossFunction):
    def __init__(
        self,
        pos_sampling_technique: Literal["all", "first"] = "all",
        neg_sampling_technique: Literal["all", "first"] = "all",
        max_num_neg_samples: int | None = None,
    ):
        super().__init__()
        self.pos_sampling_technique = pos_sampling_technique
        self.neg_sampling_technique = neg_sampling_technique
        self.max_num_neg_samples = max_num_neg_samples

    def get_ib_idcs(self, num_queries: int, num_docs: int) -> Tuple[torch.Tensor, torch.Tensor]:
        min_idx = torch.arange(num_queries)[:, None] * num_docs
        max_idx = min_idx + num_docs
        if self.pos_sampling_technique == "all":
            pos_mask = torch.arange(num_queries * num_docs)[None].greater_equal(min_idx) & torch.arange(
                num_queries * num_docs
            )[None].less(max_idx)
        elif self.pos_sampling_technique == "first":
            pos_mask = torch.arange(num_queries * num_docs)[None].eq(min_idx)
        else:
            raise ValueError("invalid pos sampling technique")
        pos_idcs = pos_mask.nonzero(as_tuple=True)[1]
        if self.neg_sampling_technique == "all":
            neg_mask = torch.arange(num_queries * num_docs)[None].less(min_idx) | torch.arange(num_queries * num_docs)[
                None
            ].greater_equal(max_idx)
        elif self.neg_sampling_technique == "first":
            neg_mask = torch.arange(num_queries * num_docs)[None, None].eq(min_idx).any(1) & torch.arange(
                num_queries * num_docs
            )[None].ne(min_idx)
        else:
            raise ValueError("invalid neg sampling technique")
        neg_idcs = neg_mask.nonzero(as_tuple=True)[1]
        if self.max_num_neg_samples is not None:
            neg_idcs = neg_idcs.view(num_queries, -1)
            if neg_idcs.shape[-1] > 1:
                neg_idcs = neg_idcs[:, torch.randperm(neg_idcs.shape[-1])]
            neg_idcs = neg_idcs[:, : self.max_num_neg_samples]
            neg_idcs = neg_idcs.reshape(-1)
        return pos_idcs, neg_idcs

    def compute_loss(self, scores: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("InBatchLossFunction.compute_loss must be implemented by subclasses")


class InBatchCrossEntropy(InBatchLossFunction):
    def compute_loss(self, scores: torch.Tensor) -> torch.Tensor:
        targets = torch.zeros(scores.shape[0], dtype=torch.long, device=scores.device)
        loss = torch.nn.functional.cross_entropy(scores, targets)
        return loss


class RegularizationLossFunction(EmbeddingLossFunction):
    def __init__(self, query_weight: float = 1e-4, doc_weight: float = 1e-4) -> None:
        self.query_weight = query_weight
        self.doc_weight = doc_weight


class L2Regularization(RegularizationLossFunction):
    def compute_loss(
        self,
        query_embeddings: torch.Tensor,
        doc_embeddings: torch.Tensor,
    ) -> torch.Tensor:
        query_loss = self.query_weight * query_embeddings.norm(dim=-1).mean()
        doc_loss = self.doc_weight * doc_embeddings.norm(dim=-1).mean()
        loss = query_loss + doc_loss
        return loss


class L1Regularization(RegularizationLossFunction):
    def compute_loss(
        self,
        query_embeddings: torch.Tensor,
        doc_embeddings: torch.Tensor,
    ) -> torch.Tensor:
        query_loss = self.query_weight * query_embeddings.norm(p=1, dim=-1).mean()
        doc_loss = self.doc_weight * doc_embeddings.norm(p=1, dim=-1).mean()
        loss = query_loss + doc_loss
        return loss


class FLOPSRegularization(RegularizationLossFunction):
    def compute_loss(
        self,
        query_embeddings: torch.Tensor,
        doc_embeddings: torch.Tensor,
    ) -> torch.Tensor:
        query_loss = torch.sum(torch.mean(torch.abs(query_embeddings), dim=0) ** 2)
        doc_loss = torch.sum(torch.mean(torch.abs(doc_embeddings), dim=0) ** 2)
        anti_zero = 1 / (torch.sum(query_embeddings) ** 2) + 1 / (torch.sum(doc_embeddings) ** 2)
        loss = self.query_weight * query_loss + self.doc_weight * doc_loss + anti_zero
        return loss



from typing import Any, Dict, Type

import pytest
import torch
from lightning import LightningDataModule, LightningModule, Trainer
from torch.utils.data import DataLoader, Dataset

from lightning_ir.lightning_utils.lr_schedulers import (
    ConstantLRSchedulerWithLinearWarmup,
    LinearLRSchedulerWithLinearWarmup,
    WarmupLRScheduler,
)
from lightning_ir.lightning_utils.schedulers import (
    GenericConstantSchedulerWithLinearWarmup,
    GenericConstantSchedulerWithQuadraticWarmup,
    GenericLinearSchedulerWithLinearWarmup,
    GenericScheduler,
)


class DummyObject:
    def __init__(self) -> None:
        self.value = 100


class DummyModule(LightningModule):
    def __init__(
        self, LRScheduler: Type[WarmupLRScheduler] | None = None, scheduler_kwargs: Dict[str, float] | None = None
    ) -> None:
        super().__init__()
        self.LRScheduler = LRScheduler
        self.scheduler_kwargs = scheduler_kwargs
        self.param = torch.nn.Parameter(torch.tensor(0.0))
        self.dummy_object = DummyObject()

    def training_step(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        return torch.tensor(0.0, requires_grad=True)

    def configure_optimizers(self):
        lr = 1
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        out = {"optimizer": optimizer}
        if self.LRScheduler is not None:
            out["lr_scheduler"] = {
                "scheduler": self.LRScheduler(optimizer, **self.scheduler_kwargs),
                "interval": "step",
            }
        return (out,)


class DummyDataset(Dataset):
    def __len__(self) -> int:
        return 100

    def __getitem__(self, index) -> Any:
        return 0


class DummyDataModule(LightningDataModule):
    def __init__(self) -> None:
        super().__init__()

    def train_dataloader(self) -> Any:
        return DataLoader(DummyDataset(), batch_size=1)


@pytest.mark.parametrize(
    "Scheduler,kwargs,intermediate_value,final_value",
    (
        (
            GenericLinearSchedulerWithLinearWarmup,
            {"num_delay_steps": 10, "num_warmup_steps": 20, "num_training_steps": 100},
            50,
            0,
        ),
        (
            GenericConstantSchedulerWithLinearWarmup,
            {"num_delay_steps": 10, "num_warmup_steps": 20},
            50,
            100,
        ),
        (
            GenericConstantSchedulerWithQuadraticWarmup,
            {"num_delay_steps": 10, "num_warmup_steps": 20},
            25,
            100,
        ),
        (
            LinearLRSchedulerWithLinearWarmup,
            {"num_delay_steps": 10, "num_warmup_steps": 20, "num_training_steps": 100},
            0.5,
            0.0,
        ),
        (
            LinearLRSchedulerWithLinearWarmup,
            {"num_delay_steps": 10, "num_warmup_steps": 20, "num_training_steps": 100, "final_value": 0.2},
            0.5,
            0.2,
        ),
        (
            ConstantLRSchedulerWithLinearWarmup,
            {"num_delay_steps": 10, "num_warmup_steps": 20},
            0.5,
            1.0,
        ),
    ),
    ids=["Linear", "Constant", "Quadratic", "LinearLR", "LinearLRFinal", "ConstantLR"],
)
def test_scheduler(
    Scheduler: Type[GenericScheduler | WarmupLRScheduler],
    kwargs: Dict[str, float],
    intermediate_value: float,
    final_value: float,
):
    if issubclass(Scheduler, WarmupLRScheduler):
        module = DummyModule(Scheduler, kwargs)
        callbacks = []
    elif issubclass(Scheduler, GenericScheduler):
        module = DummyModule()
        callbacks = [Scheduler(["dummy_object.value"], **kwargs)]
        callbacks[0].on_train_end = lambda x, y: None
    else:
        raise ValueError("Invalid scheduler")

    trainer = Trainer(
        callbacks=callbacks,
        max_steps=20,
        accumulate_grad_batches=2,
        logger=False,
        enable_checkpointing=False,
        enable_progress_bar=False,
    )
    trainer.fit(module, datamodule=DummyDataModule())

    if issubclass(Scheduler, WarmupLRScheduler):
        optimizer = module.optimizers()
        assert optimizer.param_groups[0]["lr"] == intermediate_value
    elif issubclass(Scheduler, GenericScheduler):
        assert callbacks[0].values["dummy_object.value"] == 100
        assert module.dummy_object.value == intermediate_value

    if issubclass(Scheduler, WarmupLRScheduler):
        module = DummyModule(Scheduler, kwargs)
        callbacks = []
    elif issubclass(Scheduler, GenericScheduler):
        module = DummyModule()
        callbacks = [Scheduler(["dummy_object.value"], **kwargs)]
        callbacks[0].on_train_end = lambda x, y: None
    else:
        raise ValueError("Invalid scheduler")

    trainer = Trainer(
        callbacks=callbacks,
        max_steps=100,
        accumulate_grad_batches=2,
        logger=False,
        enable_checkpointing=False,
        enable_progress_bar=False,
    )
    trainer.fit(module, datamodule=DummyDataModule())

    if issubclass(Scheduler, WarmupLRScheduler):
        optimizer = module.optimizers()
        assert optimizer.param_groups[0]["lr"] == final_value
    elif issubclass(Scheduler, GenericScheduler):
        assert callbacks[0].values["dummy_object.value"] == 100
        assert module.dummy_object.value == final_value



from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple

import torch
from lightning import LightningModule
from transformers import BatchEncoding

from ..data import RankBatch, SearchBatch, TrainBatch
from ..loss.loss import InBatchLossFunction, LossFunction
from .config import LightningIRConfig
from .model import LightningIRModel, LightningIROutput
from .tokenizer import LightningIRTokenizer
from .validation_utils import create_qrels_from_dicts, create_run_from_scores, evaluate_run


class LightningIRModule(LightningModule):
    """LightningIRModule base class. LightningIRModules contain a LightningIRModel and a LightningIRTokenizer and
    implements the training, validation, and testing steps for the model. Derived classes must implement the forward
    method for the model.
    """

    def __init__(
        self,
        model_name_or_path: str | None = None,
        config: LightningIRConfig | None = None,
        model: LightningIRModel | None = None,
        loss_functions: Sequence[LossFunction | Tuple[LossFunction, float]] | None = None,
        evaluation_metrics: Sequence[str] | None = None,
    ):
        """Initializes the LightningIRModule.

        .. _ir-measures: https://ir-measur.es/en/latest/index.html

        :param model_name_or_path: Name or path of backbone model or fine-tuned LightningIR model, defaults to None
        :type model_name_or_path: str | None, optional
        :param config: LightningIRConfig to apply when loading from backbone model, defaults to None
        :type config: LightningIRConfig | None, optional
        :param model: Already instantiated LightningIR model, defaults to None
        :type model: LightningIRModel | None, optional
        :param loss_functions: Loss functions to apply during fine-tuning, optional loss weights can be provided per
            loss function, defaults to None
        :type loss_functions: Sequence[LossFunction | Tuple[LossFunction, float]] | None, optional
        :param evaluation_metrics: Metrics corresponding to ir-measures_ measure strings to apply during validation or
            testing, defaults to None
        :type evaluation_metrics: Sequence[str] | None, optional
        :raises ValueError: If both model and model_name_or_path are provided
        :raises ValueError: If neither model nor model_name_or_path are provided
        """
        super().__init__()
        if model is not None and model_name_or_path is not None:
            raise ValueError("Only one of model or model_name_or_path must be provided.")
        if model is None:
            if model_name_or_path is None:
                raise ValueError("Either model or model_name_or_path must be provided.")
            model = LightningIRModel.from_pretrained(model_name_or_path, config=config)

        self.model: LightningIRModel = model
        self.config = self.model.config
        self.loss_functions: List[Tuple[LossFunction, float]] | None = None
        if loss_functions is not None:
            self.loss_functions = []
            for loss_function in loss_functions:
                if isinstance(loss_function, LossFunction):
                    self.loss_functions.append((loss_function, 1.0))
                else:
                    self.loss_functions.append(loss_function)
        self.evaluation_metrics = evaluation_metrics
        self.tokenizer = LightningIRTokenizer.from_pretrained(self.config.name_or_path, config=self.config)

    def on_fit_start(self) -> None:
        """Called at the very beginning of fit.

        If on DDP it is called on every process
        """
        # NOTE huggingface models are in eval mode by default
        self.train()
        return super().on_fit_start()

    def score(self, queries: Sequence[str] | str, docs: Sequence[Sequence[str]] | Sequence[str]) -> LightningIROutput:
        """Computes relevance scores for queries and documents.

        :param queries: Queries to score
        :type queries: Sequence[str]
        :param docs: Documents to score
        :type docs: Sequence[Sequence[str]]
        :return: Model output
        :rtype: LightningIROutput
        """
        if isinstance(queries, str):
            queries = (queries,)
        if isinstance(docs[0], str):
            docs = (docs,)
        batch = RankBatch(queries, docs, None, None)
        with torch.no_grad():
            return self.forward(batch)

    def forward(self, batch: TrainBatch | RankBatch | SearchBatch) -> LightningIROutput:
        """Handles the forward pass of the model.

        :param batch: Batch of training or ranking data
        :type batch: TrainBatch | RankBatch
        :raises NotImplementedError: Must be implemented by derived class
        :return: Model output
        :rtype: LightningIROutput
        """
        raise NotImplementedError

    def prepare_input(
        self, queries: Sequence[str] | None, docs: Sequence[str] | None, num_docs: Sequence[int] | int | None
    ) -> Dict[str, BatchEncoding]:
        """Tokenizes queries and documents and returns the tokenized BatchEncoding_.

        :: _BatchEncoding: https://huggingface.co/transformers/main_classes/tokenizer#transformers.BatchEncoding

        :param queries: Queries to tokenize
        :type queries: Sequence[str] | None
        :param docs: Documents to tokenize
        :type docs: Sequence[str] | None
        :param num_docs: Number of documents per query, if None num_docs is inferred by `len(docs) // len(queries)`,
            defaults to None
        :type num_docs: Sequence[int] | int | None
        :return: Tokenized queries and documents, format depends on the tokenizer
        :rtype: Dict[str, BatchEncoding]
        """
        encodings = self.tokenizer.tokenize(
            queries, docs, return_tensors="pt", padding=True, truncation=True, num_docs=num_docs
        )
        for key in encodings:
            encodings[key] = encodings[key].to(self.device)
        return encodings

    def compute_losses(self, batch: TrainBatch) -> List[torch.Tensor]:
        """Computes the losses for the batch.

        :param batch: Batch of training data
        :type batch: TrainBatch
        :raises NotImplementedError: Must be implemented by derived class
        :return: List of losses, one for each loss function
        :rtype: List[torch.Tensor]
        """
        raise NotImplementedError

    def training_step(self, batch: TrainBatch, batch_idx: int) -> torch.Tensor:
        """Handles the training step for the model.

        :param batch: Batch of training data
        :type batch: TrainBatch
        :param batch_idx: Index of the batch
        :type batch_idx: int
        :raises ValueError: If no loss functions are set
        :return: Sum of the losses weighted by the loss weights
        :rtype: torch.Tensor
        """
        if self.loss_functions is None:
            raise ValueError("Loss functions are not set")
        losses = self.compute_losses(batch)
        total_loss = torch.tensor(0)
        assert len(losses) == len(self.loss_functions)
        for (loss_function, loss_weight), loss in zip(self.loss_functions, losses):
            self.log(loss_function.__class__.__name__, loss)
            total_loss = total_loss + loss * loss_weight
        self.log("loss", total_loss, prog_bar=True)
        return total_loss

    def validation_step(
        self, batch: TrainBatch | RankBatch | SearchBatch, batch_idx: int, dataloader_idx: int = 0
    ) -> LightningIROutput:
        """Handles the validation step for the model.

        :param batch: Batch of validation or testing data
        :type batch: TrainBatch | RankBatch
        :param batch_idx: Index of the batch
        :type batch_idx: int
        :param dataloader_idx: Index of the dataloader, defaults to 0
        :type dataloader_idx: int, optional
        :return: Model output
        :rtype: LightningIROutput
        """
        output = self.forward(batch)

        if self.evaluation_metrics is None:
            return output

        dataset_id = self.get_dataset_id(dataloader_idx)
        metrics = self.validate(
            scores=output.scores,
            query_ids=batch.query_ids,
            doc_ids=batch.doc_ids,
            qrels=batch.qrels,
            targets=getattr(batch, "targets", None),
        )
        for key, value in metrics.items():
            key = f"{dataset_id}/{key}"
            self.log(key, value, batch_size=len(batch.queries))
        return output

    def test_step(
        self,
        batch: TrainBatch | RankBatch,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> LightningIROutput:
        """Handles the testing step for the model. Passes the batch to the validation step.

        :param batch: Batch of testing data
        :type batch: TrainBatch | RankBatch
        :param batch_idx: Index of the batch
        :type batch_idx: int
        :param dataloader_idx: Index of the dataloader, defaults to 0
        :type dataloader_idx: int, optional
        :return: Model output
        :rtype: LightningIROutput
        """
        return self.validation_step(batch, batch_idx, dataloader_idx)

    def get_dataset_id(self, dataloader_idx: int) -> str:
        """Gets the dataset id from the dataloader index for logging.

        .. _ir-datasets: https://ir-datasets.com/

        :param dataloader_idx: Index of the dataloader
        :type dataloader_idx: int
        :return: ir-datasets_ dataset id or dataloader index
        :rtype: str
        """
        dataset_id = str(dataloader_idx)
        datamodule = None
        try:
            datamodule = getattr(self.trainer, "datamodule", None)
            dataset_id = datamodule.inference_datasets[dataloader_idx].dataset_id
        except Exception:
            pass
        return dataset_id

    def validate(
        self,
        scores: torch.Tensor | None = None,
        query_ids: Sequence[str] | None = None,
        doc_ids: Sequence[Sequence[str]] | None = None,
        qrels: Sequence[Dict[str, int]] | None = None,
        targets: torch.Tensor | None = None,
        num_docs: Sequence[int] | None = None,
    ) -> Dict[str, float]:
        """Validates the model output with the evaluation metrics and loss functions.

        :param scores: Model output scores, defaults to None
        :type scores: torch.Tensor | None, optional
        :param query_ids: ids of the queries, defaults to None
        :type query_ids: Sequence[str] | None, optional
        :param doc_ids: ids of the documents, defaults to None
        :type doc_ids: Sequence[Sequence[str]] | None, optional
        :param qrels: Mappings of doc_id -> relevance for each query, defaults to None
        :type qrels: Sequence[Dict[str, int]] | None, optional
        :param targets: Target tensor used during fine-tuning, defaults to None
        :type targets: torch.Tensor | None, optional
        :param num_docs: Number of documents per query, defaults to None
        :type num_docs: Sequence[int] | None, optional
        :raises ValueError: If num_docs can not be parsed and query_ids are not set
        :raises ValueError: If num_docs can not be parsed and doc_ids are not set
        :return: _description_
        :rtype: Dict[str, float]
        """
        metrics: Dict[str, float] = {}
        if self.evaluation_metrics is None or scores is None:
            return metrics
        if query_ids is None:
            if num_docs is None:
                raise ValueError("num_docs must be set if query_ids is not set")
            query_ids = tuple(str(i) for i in range(len(num_docs)))
        if doc_ids is None:
            if num_docs is None:
                raise ValueError("num_docs must be set if doc_ids is not set")
            doc_ids = tuple(tuple(f"{i}-{j}" for j in range(docs)) for i, docs in enumerate(num_docs))
        metrics.update(self.validate_metrics(scores, query_ids, doc_ids, qrels))
        metrics.update(self.validate_loss(scores, query_ids, targets))
        return metrics

    def validate_metrics(
        self,
        scores: torch.Tensor,
        query_ids: Sequence[str],
        doc_ids: Sequence[Sequence[str]],
        qrels: Sequence[Dict[str, int]] | None,
    ) -> Dict[str, float]:
        """Validates the model output with the evaluation metrics.

        :param scores: Model output scores
        :type scores: torch.Tensor
        :param query_ids: ids of the queries
        :type query_ids: Sequence[str]
        :param doc_ids: ids of the documents
        :type doc_ids: Sequence[Sequence[str]]
        :param qrels: Mappings of doc_id -> relevance for each query, defaults to None
        :type qrels: Sequence[Dict[str, int]] | None
        :return: Evaluation metrics
        :rtype: Dict[str, float]
        """
        metrics: Dict[str, float] = {}
        if self.evaluation_metrics is None or qrels is None:
            return metrics
        evaluation_metrics = [metric for metric in self.evaluation_metrics if metric != "loss"]
        ir_measures_qrels = create_qrels_from_dicts(qrels)
        if evaluation_metrics and qrels is not None:
            run = create_run_from_scores(query_ids, doc_ids, scores)
            metrics.update(evaluate_run(run, ir_measures_qrels, evaluation_metrics))
        return metrics

    def validate_loss(
        self, scores: torch.Tensor, query_ids: Sequence[str], targets: torch.Tensor | None
    ) -> Dict[str, float]:
        """Validates the model output with the loss functions.

        :param scores: Model output scores
        :type scores: torch.Tensor
        :param query_ids: ids of the queries
        :type query_ids: Sequence[str]
        :param targets: Target tensor used during fine-tuning
        :type targets: torch.Tensor | None
        :return: Loss metrics
        :rtype: Dict[str, float]
        """
        metrics: Dict[str, float] = {}
        if (
            self.evaluation_metrics is None
            or "loss" not in self.evaluation_metrics
            or targets is None
            or self.loss_functions is None
        ):
            return metrics
        scores = scores.view(len(query_ids), -1)
        for loss_function, _ in self.loss_functions:
            # NOTE skip in-batch losses because they can use a lot of memory
            if isinstance(loss_function, InBatchLossFunction):
                continue
            metrics[f"validation-{loss_function.__class__.__name__}"] = loss_function.compute_loss(
                scores, targets
            ).item()
        return metrics

    def on_validation_epoch_end(self) -> None:
        """Logs the accumulated metrics for each dataloader."""
        try:
            trainer = self.trainer
        except RuntimeError:
            trainer = None
        if trainer is not None:
            metrics = trainer.callback_metrics
            accum_metrics = defaultdict(list)
            for key, value in metrics.items():
                split = key.split("/")
                if "dataloader_idx" in split[-1]:
                    accum_metrics[split[-2]].append(value)
            for key, value in accum_metrics.items():
                self.log(key, torch.stack(value).mean(), logger=False)

    def on_test_epoch_end(self) -> None:
        """Logs the accumulated metrics for each dataloader."""
        self.on_validation_epoch_end()

    def save_pretrained(self, save_path: str | Path) -> None:
        """Saves the model and tokenizer to the save path.

        :param save_path: Path to save the model and tokenizer
        :type save_path: str | Path
        """
        self.model.save_pretrained(save_path)
        self.tokenizer.save_pretrained(save_path)

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """Saves the model and tokenizer to the trainer's log directory."""
        if self.trainer is not None and self.trainer.log_dir is not None:
            if self.trainer.global_rank != 0:
                return
            _step = self.trainer.global_step
            self.config.save_step = _step
            log_dir = Path(self.trainer.log_dir)
            save_path = log_dir / "huggingface_checkpoint"
            self.save_pretrained(save_path)



from abc import ABC, abstractmethod
from typing import Any, Dict, Sequence

from lightning import Callback, LightningModule, Trainer

from ..base import LightningIRModule


class LambdaWarmupScheduler(ABC):
    def __init__(
        self,
        num_warmup_steps: int,
        num_delay_steps: int = 0,
        *args,
        **kwargs,
    ) -> None:
        self.num_warmup_steps = num_warmup_steps
        self.num_delay_steps = num_delay_steps
        super().__init__(*args, **kwargs)

    @abstractmethod
    def value_lambda(self, current_step: int) -> float: ...

    def check_delay(self, current_step: int) -> bool:
        return current_step < self.num_delay_steps

    def check_warmup(self, current_step: int) -> bool:
        return current_step < self.num_warmup_steps + self.num_delay_steps


class LinearSchedulerWithLinearWarmup(LambdaWarmupScheduler):

    def __init__(
        self,
        num_warmup_steps: int,
        num_training_steps: int,
        final_value: float = 0.0,
        num_delay_steps: int = 0,
        *args,
        **kwargs,
    ) -> None:
        self.num_training_steps = num_training_steps
        self.final_value = final_value
        super().__init__(num_warmup_steps, num_delay_steps, *args, **kwargs)

    def value_lambda(self, current_step: int) -> float:
        if self.check_delay(current_step):
            return 0.0
        if self.check_warmup(current_step):
            return (current_step - self.num_delay_steps) / self.num_warmup_steps
        current_step = current_step - self.num_delay_steps - self.num_warmup_steps
        remaining_steps = self.num_training_steps - self.num_delay_steps - self.num_warmup_steps
        step_size = (1 - self.final_value) / remaining_steps
        return max(self.final_value, 1 - step_size * current_step)


class ConstantSchedulerWithLinearWarmup(LambdaWarmupScheduler):
    def value_lambda(self, current_step: int) -> float:
        if self.check_delay(current_step):
            return 0.0
        if self.check_warmup(current_step):
            return (current_step - self.num_delay_steps) / self.num_warmup_steps
        return 1.0


class ConstantSchedulerWithQuadraticWarmup(LambdaWarmupScheduler):
    def value_lambda(self, current_step: int) -> float:
        if self.check_delay(current_step):
            return 0.0
        if self.check_warmup(current_step):
            return ((current_step - self.num_delay_steps) / self.num_warmup_steps) ** 2
        return 1.0


class GenericScheduler(Callback, ABC):

    def __init__(self, keys: Sequence[str], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.keys = keys
        self.values: Dict[str, float] = {}

    @abstractmethod
    def step(self, key: str, current_step: int) -> float: ...

    def get_value(self, sub_keys: Sequence[str], obj: object) -> object:
        for sub_key in sub_keys:
            try:
                obj = obj[int(sub_key)]
            except ValueError:
                obj = getattr(obj, sub_key)
        return obj

    def set_value(self, sub_keys: Sequence[str], obj: object, value: float) -> None:
        obj = self.get_value(sub_keys[:-1], obj)
        setattr(obj, sub_keys[-1], value)

    def on_train_start(self, trainer: Trainer, pl_module: LightningIRModule) -> None:
        for key in self.keys:
            sub_keys = key.split(".")
            self.values[key] = float(self.get_value(sub_keys, pl_module))

    def on_train_batch_start(self, trainer: Trainer, pl_module: LightningModule, batch: Any, batch_idx: int) -> None:
        step = trainer.global_step + 1
        for key in self.keys:
            value = self.step(key, step)
            sub_keys = key.split(".")
            self.set_value(sub_keys, pl_module, value)

    def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        for key in self.keys:
            value = self.values[key]
            sub_keys = key.split(".")
            self.set_value(sub_keys, pl_module, value)


class GenericLinearSchedulerWithLinearWarmup(GenericScheduler, LinearSchedulerWithLinearWarmup):
    def step(self, key: str, current_step: int) -> float:
        value = self.values[key]
        return value * self.value_lambda(current_step)


class GenericConstantSchedulerWithLinearWarmup(GenericScheduler, ConstantSchedulerWithLinearWarmup):
    def step(self, key: str, current_step: int) -> float:
        value = self.values[key]
        return value * self.value_lambda(current_step)


class GenericConstantSchedulerWithQuadraticWarmup(GenericScheduler, ConstantSchedulerWithQuadraticWarmup):
    def step(self, key: str, current_step: int) -> float:
        value = self.values[key]
        return value * self.value_lambda(current_step)



from .config import CrossEncoderConfig
from .model import CrossEncoderModel, CrossEncoderOutput
from .module import CrossEncoderModule
from .tokenizer import CrossEncoderTokenizer

__all__ = [
    "CrossEncoderConfig",
    "CrossEncoderModel",
    "CrossEncoderModule",
    "CrossEncoderOutput",
    "CrossEncoderTokenizer",
]



from .data import DocSample, IndexBatch, QuerySample, RankBatch, RankSample, SearchBatch, TrainBatch
from .datamodule import LightningIRDataModule
from .dataset import DocDataset, QueryDataset, RunDataset, TupleDataset

__all__ = [
    "DocDataset",
    "DocSample",
    "IndexBatch",
    "LightningIRDataModule",
    "QueryDataset",
    "QuerySample",
    "RankBatch",
    "RunDataset",
    "RankSample",
    "SearchBatch",
    "TrainBatch",
    "TupleDataset",
]



import warnings
from typing import Dict, Sequence, Type

from transformers import TOKENIZER_MAPPING, BatchEncoding

from .class_factory import LightningIRTokenizerClassFactory
from .config import LightningIRConfig
from .external_model_hub import CHECKPOINT_MAPPING


class LightningIRTokenizer:
    """Base class for LightningIR tokenizers. Derived classes implement the tokenize method for handling query
    and document tokenization. It acts as mixin for a transformers.PreTrainedTokenizer_ backbone tokenizer.

    .. _transformers.PreTrainedTokenizer: \
https://huggingface.co/transformers/main_classes/tokenizer.htmltransformers.PreTrainedTokenizer
    """

    config_class: Type[LightningIRConfig] = LightningIRConfig
    """Configuration class for the tokenizer."""

    def __init__(self, *args, query_length: int = 32, doc_length: int = 512, **kwargs):
        """Initializes the tokenizer.

        :param query_length: Maximum number of tokens per query, defaults to 32
        :type query_length: int, optional
        :param doc_length: Maximum number of tokens per document, defaults to 512
        :type doc_length: int, optional
        """
        super().__init__(*args, **kwargs)
        self.query_length = query_length
        self.doc_length = doc_length

    def tokenize(
        self, queries: str | Sequence[str] | None = None, docs: str | Sequence[str] | None = None, **kwargs
    ) -> Dict[str, BatchEncoding]:
        """Tokenizes queries and documents.

        :param queries: Queries to tokenize, defaults to None
        :type queries: str | Sequence[str] | None, optional
        :param docs: Documents to tokenize, defaults to None
        :type docs: str | Sequence[str] | None, optional
        :raises NotImplementedError: Must be implemented by the derived class
        :return: Dictionary of tokenized queries and documents
        :rtype: Dict[str, BatchEncoding]
        """
        raise NotImplementedError

    @classmethod
    def from_pretrained(cls, model_name_or_path: str, *args, **kwargs) -> "LightningIRTokenizer":
        """Loads a pretrained tokenizer. Wraps the transformers.PreTrainedTokenizer.from_pretrained_ method to return a
        derived LightningIRTokenizer class. See :class:`.LightningIRTokenizerClassFactory` for more details.

        .. _transformers.PreTrainedTokenizer.from_pretrained: \
https://huggingface.co/docs/transformers/main_classes/tokenizer.html#transformers.PreTrainedTokenizer.from_pretrained

        .. highlight:: python
        .. code-block:: python

            >>> Loading using model class and backbone checkpoint
            >>> type(BiEncoderTokenizer.from_pretrained("bert-base-uncased"))
            ...
            <class 'lightning_ir.base.class_factory.BiEncoderBertTokenizerFast'>
            >>> Loading using base class and backbone checkpoint
            >>> type(LightningIRTokenizer.from_pretrained("bert-base-uncased", config=BiEncoderConfig()))
            ...
            <class 'lightning_ir.base.class_factory.BiEncoderBertTokenizerFast'>

        :param model_name_or_path: Name or path of the pretrained tokenizer
        :type model_name_or_path: str
        :raises ValueError: If called on the abstract class :class:`LightningIRTokenizer` and no config is passed
        :return: A derived LightningIRTokenizer consisting of a backbone tokenizer and a LightningIRTokenizer mixin
        :rtype: LightningIRTokenizer
        """
        config = kwargs.pop("config", None)
        if config is not None:
            kwargs.update(config.to_tokenizer_dict())
        if cls is LightningIRTokenizer or all(issubclass(base, LightningIRTokenizer) for base in cls.__bases__):
            # no backbone models found, create derived lightning-ir tokenizer based on backbone model
            if model_name_or_path in CHECKPOINT_MAPPING:
                _config = CHECKPOINT_MAPPING[model_name_or_path]
                Config = _config.__class__
                if config is not None:
                    warnings.warn(f"{model_name_or_path} is a registered checkpoint. The provided config is ignored.")
                kwargs.update(_config.to_tokenizer_dict())
            elif config is not None:
                Config = config.__class__
            elif cls is not LightningIRTokenizer and hasattr(cls, "config_class"):
                Config = cls.config_class
            else:
                Config = LightningIRTokenizerClassFactory.get_lightning_ir_config(model_name_or_path)
                if Config is None:
                    raise ValueError("Pass a config to `from_pretrained`.")
            BackboneConfig = LightningIRTokenizerClassFactory.get_backbone_config(model_name_or_path)
            BackboneTokenizers = TOKENIZER_MAPPING[BackboneConfig]
            if kwargs.get("use_fast", True):
                BackboneTokenizer = BackboneTokenizers[1]
            else:
                BackboneTokenizer = BackboneTokenizers[0]
            cls = LightningIRTokenizerClassFactory(Config).from_backbone_class(BackboneTokenizer)
            return cls.from_pretrained(model_name_or_path, *args, **kwargs)
        return super(LightningIRTokenizer, cls).from_pretrained(model_name_or_path, *args, **kwargs)



from typing import Sequence

import pytest
import torch
from _pytest.fixtures import SubRequest

from lightning_ir.bi_encoder.config import BiEncoderConfig
from lightning_ir.cross_encoder.model import CrossEncoderConfig
from lightning_ir.data.data import IndexBatch, SearchBatch, TrainBatch
from lightning_ir.data.datamodule import LightningIRDataModule
from lightning_ir.data.dataset import RunDataset, TupleDataset

from .conftest import RUNS_DIR


@pytest.fixture(params=[BiEncoderConfig(), CrossEncoderConfig()], ids=["BiEncoder", "CrossEncoder"])
def rank_run_datamodule(
    model_name_or_path: str,
    inference_datasets: Sequence[RunDataset],
    request: SubRequest,
) -> LightningIRDataModule:
    train_dataset = RunDataset(
        RUNS_DIR / "lightning-ir.tsv",
        depth=5,
        sample_size=2,
        sampling_strategy="top",
        targets="rank",
    )
    datamodule = LightningIRDataModule(
        model_name_or_path=model_name_or_path,
        config=request.param,
        num_workers=0,
        train_batch_size=2,
        inference_batch_size=2,
        train_dataset=train_dataset,
        inference_datasets=inference_datasets,
    )
    datamodule.setup(stage="fit")
    return datamodule


@pytest.fixture(params=[BiEncoderConfig(), CrossEncoderConfig()], ids=["BiEncoder", "CrossEncoder"])
def relevance_run_datamodule(
    model_name_or_path: str,
    inference_datasets: Sequence[RunDataset],
    request: SubRequest,
) -> LightningIRDataModule:
    train_dataset = RunDataset(
        RUNS_DIR / "lightning-ir.tsv",
        depth=5,
        sample_size=2,
        sampling_strategy="top",
        targets="relevance",
    )
    datamodule = LightningIRDataModule(
        model_name_or_path=model_name_or_path,
        config=request.param,
        num_workers=0,
        train_batch_size=2,
        inference_batch_size=2,
        train_dataset=train_dataset,
        inference_datasets=inference_datasets,
    )
    datamodule.setup(stage="fit")
    return datamodule


@pytest.fixture(params=[BiEncoderConfig(), CrossEncoderConfig()], ids=["BiEncoder", "CrossEncoder"])
def single_relevant_run_datamodule(
    model_name_or_path: str,
    inference_datasets: Sequence[RunDataset],
    request: SubRequest,
) -> LightningIRDataModule:
    train_dataset = RunDataset(
        RUNS_DIR / "lightning-ir.tsv",
        depth=5,
        sample_size=2,
        sampling_strategy="single_relevant",
        targets="relevance",
    )
    datamodule = LightningIRDataModule(
        model_name_or_path=model_name_or_path,
        config=request.param,
        num_workers=0,
        train_batch_size=2,
        inference_batch_size=2,
        train_dataset=train_dataset,
        inference_datasets=inference_datasets,
    )
    datamodule.setup(stage="fit")
    return datamodule


@pytest.fixture(params=[BiEncoderConfig(), CrossEncoderConfig()], ids=["BiEncoder", "CrossEncoder"])
def tuples_datamodule(
    model_name_or_path: str,
    inference_datasets: Sequence[RunDataset],
    request: SubRequest,
) -> LightningIRDataModule:
    train_dataset = TupleDataset("lightning-ir", targets="order", num_docs=2)
    datamodule = LightningIRDataModule(
        model_name_or_path=model_name_or_path,
        config=request.param,
        num_workers=0,
        train_batch_size=2,
        inference_batch_size=2,
        train_dataset=train_dataset,
        inference_datasets=inference_datasets,
    )
    datamodule.setup(stage="fit")
    return datamodule


def test_rank_run_dataset(rank_run_datamodule: LightningIRDataModule):
    datamodule = rank_run_datamodule
    dataloader = datamodule.train_dataloader()
    dataset = dataloader.dataset

    batch = next(iter(dataloader))
    assert isinstance(batch, TrainBatch)
    for value in batch.__dict__.values():
        assert value is not None
    assert batch.targets.shape[0] == datamodule.train_batch_size * dataset.sample_size
    target_ranks = dataset.depth - torch.arange(dataset.sample_size).repeat(datamodule.train_batch_size)
    assert (batch.targets[..., 0] == target_ranks).all()


def test_relevance_run_dataset(relevance_run_datamodule: LightningIRDataModule):
    datamodule = relevance_run_datamodule
    dataloader = datamodule.train_dataloader()
    dataset = dataloader.dataset

    batch = next(iter(dataloader))
    assert isinstance(batch, TrainBatch)
    for value in batch.__dict__.values():
        assert value is not None
    assert batch.targets.shape[0] == datamodule.train_batch_size * dataset.sample_size


def test_single_relevant_run_dataset(
    single_relevant_run_datamodule: LightningIRDataModule,
):
    datamodule = single_relevant_run_datamodule
    dataloader = datamodule.train_dataloader()

    batch = next(iter(dataloader))
    assert isinstance(batch, TrainBatch)
    for value in batch.__dict__.values():
        assert value is not None
    assert (batch.targets.max(-1).values > 0).sum() == datamodule.train_batch_size


def test_tuples_dataset(tuples_datamodule: LightningIRDataModule):
    dataloader = tuples_datamodule.train_dataloader()
    dataset = dataloader.dataset

    batch = next(iter(dataloader))
    assert isinstance(batch, TrainBatch)
    for field in batch.__dict__.keys():
        value = getattr(batch, field)
        if field == "qrels":
            assert value is None
        else:
            assert value is not None
    assert batch.targets.shape[0] == dataloader.batch_size * dataset.num_docs


def test_query_dataset(query_datamodule: LightningIRDataModule):
    dataloader = query_datamodule.test_dataloader()[0]
    batch: SearchBatch = next(iter(dataloader))
    assert isinstance(batch, SearchBatch)
    for field in batch.__dict__.keys():
        value = getattr(batch, field)
        if field in ("queries", "query_ids", "qrels"):
            assert value is not None
        else:
            assert value is None


def test_doc_dataset(doc_datamodule: LightningIRDataModule):
    dataloader = doc_datamodule.test_dataloader()[0]
    batch: IndexBatch = next(iter(dataloader))
    assert isinstance(batch, IndexBatch)
    for field in batch.__dict__.keys():
        value = getattr(batch, field)
        if field in ("docs", "doc_ids"):
            assert value is not None
        else:
            assert value is None


def test_json_dataset():
    dataset = RunDataset(
        RUNS_DIR / "run.jsonl",
        depth=5,
        sample_size=5,
        sampling_strategy="top",
        targets="rank",
    )
    sample = dataset[0]
    assert sample is not None
    assert sample.query_id is not None
    assert len(sample.doc_ids) == 5



from .faiss_indexer import (
    FaissFlatIndexConfig,
    FaissFlatIndexer,
    FaissIVFIndexConfig,
    FaissIVFIndexer,
    FaissIVFPQIndexConfig,
    FaissIVFPQIndexer,
)
from .faiss_searcher import FaissSearchConfig, FaissSearcher
from .indexer import IndexConfig, Indexer
from .searcher import SearchConfig, Searcher
from .sparse_indexer import SparseIndexConfig, SparseIndexer
from .sparse_searcher import SparseSearchConfig, SparseSearcher

__all__ = [
    "FaissFlatIndexConfig",
    "FaissFlatIndexer",
    "FaissIVFIndexConfig",
    "FaissIVFIndexer",
    "FaissIVFPQIndexConfig",
    "FaissIVFPQIndexer",
    "FaissSearchConfig",
    "FaissSearcher",
    "IndexConfig",
    "Indexer",
    "SearchConfig",
    "Searcher",
    "SparseIndexConfig",
    "SparseIndexer",
    "SparseSearcher",
    "SparseSearchConfig",
]



from .callbacks import IndexCallback, RankCallback, ReRankCallback, SearchCallback
from .lr_schedulers import (
    LR_SCHEDULERS,
    ConstantLRSchedulerWithLinearWarmup,
    LinearLRSchedulerWithLinearWarmup,
    WarmupLRScheduler,
)
from .schedulers import (
    GenericConstantSchedulerWithLinearWarmup,
    GenericConstantSchedulerWithQuadraticWarmup,
    GenericLinearSchedulerWithLinearWarmup,
)

__all__ = [
    "ConstantLRSchedulerWithLinearWarmup",
    "GenericConstantSchedulerWithLinearWarmup",
    "GenericConstantSchedulerWithQuadraticWarmup",
    "GenericLinearSchedulerWithLinearWarmup",
    "IndexCallback",
    "LinearLRSchedulerWithLinearWarmup",
    "LR_SCHEDULERS",
    "RankCallback",
    "ReRankCallback",
    "SearchCallback",
    "WarmupLRScheduler",
]



from typing import Type

import pytest
import torch

from lightning_ir.loss.loss import (
    ApproxMRR,
    ApproxNDCG,
    ApproxRankMSE,
    ConstantMarginMSE,
    FLOPSRegularization,
    InBatchCrossEntropy,
    InBatchLossFunction,
    KLDivergence,
    L1Regularization,
    L2Regularization,
    LocalizedContrastiveEstimation,
    RankNet,
    RegularizationLossFunction,
    ScoringLossFunction,
    SupervisedMarginMSE,
)

torch.manual_seed(42)


@pytest.fixture(scope="module")
def batch_size() -> int:
    return 4


@pytest.fixture(scope="module")
def sequence_length() -> int:
    return 8


@pytest.fixture(scope="module")
def depth() -> int:
    return 10


@pytest.fixture(scope="module")
def embedding_dim() -> int:
    return 4


@pytest.fixture(scope="module")
def scores(batch_size: int, depth: int) -> torch.Tensor:
    tensor = torch.randn((batch_size, depth), requires_grad=True)
    return tensor


@pytest.fixture(scope="module")
def labels(batch_size: int, depth: int) -> torch.Tensor:
    tensor = torch.randint(0, 5, (batch_size, depth))
    return tensor


@pytest.fixture(scope="module")
def embeddings(batch_size: int, sequence_length: int, embedding_dim: int) -> torch.Tensor:
    tensor = torch.randn((batch_size, sequence_length, embedding_dim), requires_grad=True)
    return tensor


@pytest.mark.parametrize(
    "LossFunc",
    [
        ApproxMRR,
        ApproxNDCG,
        ApproxRankMSE,
        ConstantMarginMSE,
        KLDivergence,
        LocalizedContrastiveEstimation,
        RankNet,
        SupervisedMarginMSE,
    ],
)
def test_loss_func(
    scores: torch.Tensor,
    labels: torch.Tensor,
    LossFunc: Type[ScoringLossFunction],
):
    loss_func = LossFunc()
    loss = loss_func.compute_loss(scores, labels)
    assert loss >= 0
    assert loss.requires_grad


@pytest.mark.parametrize(
    "InBatchLossFunc",
    [InBatchCrossEntropy],
)
def test_in_batch_loss_func(InBatchLossFunc: Type[InBatchLossFunction], scores: torch.Tensor):
    loss_func = InBatchLossFunc()
    loss = loss_func.compute_loss(scores)
    assert loss >= 0
    assert loss.requires_grad


@pytest.mark.parametrize("RegularizationLossFunc", [L1Regularization, L2Regularization, FLOPSRegularization])
def test_regularization_loss_func(
    RegularizationLossFunc: Type[RegularizationLossFunction],
    embeddings: torch.Tensor,
):
    loss_func = RegularizationLossFunc()
    loss = loss_func.compute_loss(embeddings, embeddings)
    assert loss >= 0
    assert loss.requires_grad



from typing import Dict, Sequence

import ir_measures
import numpy as np
import pandas as pd
import torch


def create_run_from_scores(
    query_ids: Sequence[str], doc_ids: Sequence[Sequence[str]], scores: torch.Tensor
) -> pd.DataFrame:
    """Convience function to create a run DataFrame from query and document ids and scores.

    :param query_ids: Query ids
    :type query_ids: Sequence[str]
    :param doc_ids: Document ids
    :type doc_ids: Sequence[Sequence[str]]
    :param scores: Scores
    :type scores: torch.Tensor
    :return: DataFrame with query_id, q0, doc_id, score, rank, and system columns
    :rtype: pd.DataFrame
    """
    num_docs = [len(ids) for ids in doc_ids]
    scores = scores.float().cpu().detach().numpy().reshape(-1)
    df = pd.DataFrame(
        {
            "query_id": np.array(query_ids).repeat(num_docs),
            "q0": 0,
            "doc_id": np.array(sum(map(lambda x: list(x), doc_ids), [])),
            "score": scores,
            "system": "lightning_ir",
        }
    )
    df["rank"] = df.groupby("query_id")["score"].rank(ascending=False, method="first")

    def key(series: pd.Series) -> pd.Series:
        if series.name == "query_id":
            return series.map({query_id: i for i, query_id in enumerate(query_ids)})
        return series

    df = df.sort_values(["query_id", "rank"], ascending=[True, True], key=key)
    return df


def create_qrels_from_dicts(qrels: Sequence[Dict[str, int]]) -> pd.DataFrame:
    """Convience function to create a qrels DataFrame from a list of dictionaries.

    :param qrels: Mappings of doc_id -> relevance for each query, defaults to None
    :type qrels: Sequence[Dict[str, int]]
    :return: DataFrame with query_id, q0, doc_id, and relevance columns
    :rtype: pd.DataFrame
    """
    return pd.DataFrame.from_records(qrels)


def evaluate_run(run: pd.DataFrame, qrels: pd.DataFrame, measures: Sequence[str]) -> Dict[str, float]:
    """Convience function to evaluate a run against qrels using a set of measures.

    .. _ir-measures: https://ir-measur.es/en/latest/index.html

    :param run: Parse TREC run
    :type run: pd.DataFrame
    :param qrels: Parse TREC qrels
    :type qrels: pd.DataFrame
    :param measures: Metrics corresponding to ir-measures_ measure strings
    :type measures: Sequence[str]
    :return: Calculated metrics
    :rtype: Dict[str, float]
    """
    parsed_measures = [ir_measures.parse_measure(measure) for measure in measures]
    metrics = {str(measure): measure.calc_aggregate(qrels, run) for measure in parsed_measures}
    return metrics



from __future__ import annotations

import itertools
from dataclasses import is_dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Sequence, Tuple, TypeVar

import pandas as pd
import torch
from lightning import LightningModule, Trainer
from lightning.pytorch.callbacks import BasePredictionWriter, Callback, TQDMProgressBar

from ..data import RankBatch, SearchBatch
from ..data.dataset import RUN_HEADER, DocDataset, QueryDataset, RunDataset
from ..retrieve import IndexConfig, Indexer, SearchConfig, Searcher

if TYPE_CHECKING:
    from ..base import LightningIRModule, LightningIROutput
    from ..bi_encoder import BiEncoderModule, BiEncoderOutput

T = TypeVar("T")


def format_large_number(number: float) -> str:
    suffixes = ["", "K", "M", "B", "T"]
    suffix_index = 0

    while number >= 1000 and suffix_index < len(suffixes) - 1:
        number /= 1000.0
        suffix_index += 1

    formatted_number = "{:.2f}".format(number)

    suffix = suffixes[suffix_index]
    if suffix:
        formatted_number += f" {suffix}"
    return formatted_number


class GatherMixin:
    def gather(self, pl_module: LightningIRModule, dataclass: T) -> T:
        if is_dataclass(dataclass):
            return dataclass.__class__(
                **{k: self.gather(pl_module, getattr(dataclass, k)) for k in dataclass.__dataclass_fields__}
            )
        return pl_module.all_gather(dataclass)


class IndexCallback(Callback, GatherMixin):
    def __init__(
        self,
        index_dir: Path | str | None,
        index_config: IndexConfig,
        verbose: bool = False,
    ) -> None:
        super().__init__()
        self.index_config = index_config
        self.index_dir = index_dir
        self.verbose = verbose
        self.indexer: Indexer

    def setup(self, trainer: Trainer, pl_module: BiEncoderModule, stage: str) -> None:
        if stage != "test":
            raise ValueError("IndexCallback can only be used in test stage")

    def on_test_start(self, trainer: Trainer, pl_module: BiEncoderModule) -> None:
        dataloaders = trainer.test_dataloaders
        if dataloaders is None:
            raise ValueError("No test_dataloaders found")
        datasets = [dataloader.dataset for dataloader in dataloaders]
        if not all(isinstance(dataset, DocDataset) for dataset in datasets):
            raise ValueError("Expected DocDatasets for indexing")

    def get_index_dir(self, pl_module: BiEncoderModule, dataset: DocDataset) -> Path:
        index_dir = self.index_dir
        if index_dir is None:
            default_index_dir = Path(pl_module.config.name_or_path)
            if default_index_dir.exists():
                index_dir = default_index_dir / "indexes"
            else:
                raise ValueError("No index_dir provided and model_name_or_path is not a path")
            index_dir = index_dir / dataset.docs_dataset_id
        return Path(index_dir)

    def get_indexer(self, trainer: Trainer, pl_module: BiEncoderModule, dataset_idx: int) -> Indexer:
        dataloaders = trainer.test_dataloaders
        if dataloaders is None:
            raise ValueError("No test_dataloaders found")
        dataset = dataloaders[dataset_idx].dataset

        index_dir = self.get_index_dir(pl_module, dataset)

        indexer = self.index_config.indexer_class(index_dir, self.index_config, pl_module.config, self.verbose)
        return indexer

    def log_to_pg(self, info: Dict[str, Any], trainer: Trainer):
        pg_callback = trainer.progress_bar_callback
        if pg_callback is None or not isinstance(pg_callback, TQDMProgressBar):
            return
        pg = pg_callback.test_progress_bar
        info = {k: format_large_number(v) for k, v in info.items()}
        if pg is not None:
            pg.set_postfix(info)

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: BiEncoderModule,
        outputs: BiEncoderOutput,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if batch_idx == 0:
            if hasattr(self, "indexer"):
                self.indexer.save()
            self.indexer = self.get_indexer(trainer, pl_module, dataloader_idx)

        batch = self.gather(pl_module, batch)
        outputs = self.gather(pl_module, outputs)

        if not trainer.is_global_zero:
            return

        self.indexer.add(batch, outputs)
        self.log_to_pg(
            {
                "num_docs": self.indexer.num_docs,
                "num_embeddings": self.indexer.num_embeddings,
            },
            trainer,
        )
        return super().on_test_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)

    def on_test_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.indexer.save()


class RankCallback(BasePredictionWriter, GatherMixin):
    def __init__(self, save_dir: Path | str | None = None, run_name: str | None = None) -> None:
        super().__init__()
        self.save_dir = Path(save_dir) if save_dir is not None else None
        self.run_name = run_name
        self.run_dfs: List[pd.DataFrame] = []

    def setup(self, trainer: Trainer, pl_module: LightningIRModule, stage: str) -> None:
        if stage != "test":
            raise ValueError(f"{self.__class__.__name__} can only be used in test stage")

    def on_test_epoch_start(self, trainer: Trainer, pl_module: LightningIRModule) -> None:
        super().on_test_epoch_start(trainer, pl_module)
        self.run_dfs = []
        if self.save_dir is None:
            default_save_dir = Path(pl_module.config.name_or_path)
            if default_save_dir.exists():
                self.save_dir = default_save_dir / "runs"
                print(f"Using default save_dir {self.save_dir}")
            else:
                raise ValueError("No save_dir provided and model_name_or_path is not a path")

    def on_test_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_test_epoch_end(trainer, pl_module)
        if trainer.is_global_zero:
            self.write_run_dfs(trainer, -1)
            self.run_dfs = []

    def get_run_path(self, trainer: Trainer, dataset_idx: int) -> Path:
        dataloaders = trainer.test_dataloaders
        if self.save_dir is None:
            raise ValueError("No save_dir found; call setup before using this method")
        if dataloaders is None:
            raise ValueError("No test_dataloaders found")
        dataset = dataloaders[dataset_idx].dataset
        if self.run_name is not None:
            run_file = self.run_name
        elif isinstance(dataset, QueryDataset):
            run_file = f"{dataset.dataset_id.replace('/', '-')}.run"
        elif isinstance(dataset, RunDataset):
            if dataset.run_path is None:
                run_file = f"{dataset.dataset_id.replace('/', '-')}.run"
            else:
                run_file = f"{dataset.run_path.name.split('.')[0]}.run"
        run_file_path = self.save_dir / run_file
        return run_file_path

    def rank(self, batch: RankBatch, output: LightningIROutput) -> Tuple[torch.Tensor, List[str], List[int]]:
        scores = output.scores
        if scores is None:
            raise ValueError("Expected output to have scores")
        doc_ids = batch.doc_ids
        if doc_ids is None:
            raise ValueError("Expected batch to have doc_ids")
        scores = scores.view(-1)
        num_docs = [len(_doc_ids) for _doc_ids in doc_ids]
        doc_ids = list(itertools.chain.from_iterable(doc_ids))
        if scores.shape[0] != len(doc_ids):
            raise ValueError("scores and doc_ids must have the same length")
        return scores, doc_ids, num_docs

    def write_run_dfs(self, trainer: Trainer, dataloader_idx: int):
        if not trainer.is_global_zero or not self.run_dfs:
            return
        run_file_path = self.get_run_path(trainer, dataloader_idx)
        run_file_path.parent.mkdir(parents=True, exist_ok=True)
        run_df = pd.concat(self.run_dfs, ignore_index=True)
        run_df.to_csv(run_file_path, header=False, index=False, sep="\t")

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningIRModule,
        outputs: LightningIROutput,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        batch_indices = trainer.predict_loop.current_batch_indices
        self.write_on_batch_end(trainer, pl_module, outputs, batch_indices, batch, batch_idx, dataloader_idx)
        super().on_test_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)

    def write_on_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningIRModule,
        prediction: LightningIROutput,
        batch_indices: Sequence[int] | None,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        batch = self.gather(pl_module, batch)
        prediction = self.gather(pl_module, prediction)
        if not trainer.is_global_zero:
            return

        query_ids = batch.query_ids
        if query_ids is None:
            raise ValueError("Expected batch to have query_ids")
        scores, doc_ids, num_docs = self.rank(batch, prediction)
        scores = scores.float().cpu().numpy()

        query_ids = list(
            itertools.chain.from_iterable(itertools.repeat(query_id, num) for query_id, num in zip(query_ids, num_docs))
        )
        run_df = pd.DataFrame(zip(query_ids, doc_ids, scores), columns=["query_id", "doc_id", "score"])
        run_df = run_df.sort_values(["query_id", "score"], ascending=[True, False])
        run_df["rank"] = run_df.groupby("query_id")["score"].rank(ascending=False, method="first").astype(int)
        run_df["q0"] = 0
        run_df["system"] = pl_module.model.__class__.__name__
        run_df = run_df[RUN_HEADER]

        if batch_idx == 0:
            self.write_run_dfs(trainer, dataloader_idx - 1)
            self.run_dfs = []
        self.run_dfs.append(run_df)


class SearchCallback(RankCallback):
    def __init__(
        self,
        index_dir: Path | str,
        search_config: SearchConfig,
        save_dir: Path | str | None = None,
        use_gpu: bool = True,
    ) -> None:
        super().__init__(save_dir)
        self.index_dir = index_dir
        self.search_config = search_config
        self.use_gpu = use_gpu
        self.searcher: Searcher

    def setup(self, trainer: Trainer, pl_module: BiEncoderModule, stage: str) -> None:
        if stage != "test":
            raise ValueError(f"{self.__class__.__name__} can only be used in test stage")
        self.searcher = self.search_config.search_class(self.index_dir, self.search_config, pl_module, self.use_gpu)
        pl_module.searcher = self.searcher

    def rank(
        self, batch: SearchBatch | RankBatch, output: LightningIROutput
    ) -> Tuple[torch.Tensor | List[str] | List[int]]:
        if isinstance(batch, SearchBatch):
            doc_scores, flat_doc_ids, num_docs = self.searcher.search(output)
            cum_num_docs = [0] + [sum(num_docs[: i + 1]) for i in range(len(num_docs))]
            doc_ids = tuple(tuple(flat_doc_ids[cum_num_docs[i] : cum_num_docs[i + 1]]) for i in range(len(num_docs)))
            output.scores = doc_scores
            dummy_docs = [[""] * num for num in num_docs]
            batch = RankBatch(batch.queries, dummy_docs, batch.query_ids, doc_ids, batch.qrels)
        return super().rank(batch, output)


class ReRankCallback(RankCallback):
    pass



from . import flash_bert

FLASH_ATTENTION_MAP = {
    "bert": (flash_bert.flash_attention_forward, flash_bert.SELF_ATTENTION_PATTERN),
    "roberta": (flash_bert.flash_attention_forward, flash_bert.SELF_ATTENTION_PATTERN),
    "electra": (flash_bert.flash_attention_forward, flash_bert.SELF_ATTENTION_PATTERN),
}



from typing import Literal, Sequence

from ..bi_encoder import BiEncoderConfig, BiEncoderModel


class ColConfig(BiEncoderConfig):
    model_type = "col"

    def __init__(
        self,
        query_expansion: bool = True,
        doc_mask_scoring_tokens: Sequence[str] | Literal["punctuation"] | None = "punctuation",
        embedding_dim: int = 128,
        projection: Literal["linear", "linear_no_bias"] | None = "linear_no_bias",
        **kwargs,
    ) -> None:
        kwargs["query_pooling_strategy"] = None
        kwargs["doc_expansion"] = False
        kwargs["attend_to_doc_expanded_tokens"] = False
        kwargs["doc_pooling_strategy"] = None
        super().__init__(
            query_expansion=query_expansion,
            doc_mask_scoring_tokens=doc_mask_scoring_tokens,
            embedding_dim=embedding_dim,
            projection=projection,
            **kwargs,
        )


class ColModel(BiEncoderModel):
    config_class = ColConfig



from .config import BiEncoderConfig
from .model import BiEncoderEmbedding, BiEncoderModel, BiEncoderOutput, ScoringFunction
from .module import BiEncoderModule
from .tokenizer import BiEncoderTokenizer

__all__ = [
    "BiEncoderConfig",
    "BiEncoderEmbedding",
    "BiEncoderModel",
    "BiEncoderModule",
    "BiEncoderOutput",
    "BiEncoderTokenizer",
    "ScoringFunction",
]



from dataclasses import dataclass
from typing import Any, Dict, List, Sequence

import torch


@dataclass
class RankSample:
    query_id: str
    query: str
    doc_ids: Sequence[str]
    docs: Sequence[str]
    targets: torch.Tensor | None = None
    qrels: List[Dict[str, Any]] | None = None


@dataclass
class QuerySample:
    query_id: str
    query: str

    @classmethod
    def from_ir_dataset_sample(cls, sample):
        return cls(sample[0], sample[1])


@dataclass
class DocSample:
    doc_id: str
    doc: str

    @classmethod
    def from_ir_dataset_sample(cls, sample):
        return cls(sample[0], sample.default_text())


@dataclass
class RankBatch:
    queries: Sequence[str]
    docs: Sequence[Sequence[str]]
    query_ids: Sequence[str] | None = None
    doc_ids: Sequence[Sequence[str]] | None = None
    qrels: List[Dict[str, int]] | None = None


@dataclass
class TrainBatch(RankBatch):
    targets: torch.Tensor | None = None


@dataclass
class IndexBatch:
    doc_ids: Sequence[str]
    docs: Sequence[str]


@dataclass
class SearchBatch:
    query_ids: Sequence[str]
    queries: Sequence[str]
    doc_ids: Sequence[Sequence[str]] | None = None
    qrels: List[Dict[str, int]] | None = None



import os
import sys
from pathlib import Path
from typing import Any, Dict, List, Mapping, Set

import torch
from lightning import LightningDataModule, LightningModule, Trainer
from lightning.fabric.loggers.logger import _DummyExperiment as DummyExperiment
from lightning.pytorch.cli import LightningCLI, SaveConfigCallback
from lightning.pytorch.loggers import WandbLogger
from typing_extensions import override

import lightning_ir  # noqa: F401
from lightning_ir.lightning_utils.lr_schedulers import LR_SCHEDULERS, WarmupLRScheduler

if torch.cuda.is_available():
    torch.set_float32_matmul_precision("medium")

sys.path.append(str(Path.cwd()))

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class LightningIRSaveConfigCallback(SaveConfigCallback):
    @override
    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if stage != "fit" or trainer.logger is None:
            return
        return super().setup(trainer, pl_module, stage)


class LightningIRWandbLogger(WandbLogger):
    @property
    def save_dir(self) -> str | None:
        """Gets the save directory.

        Returns:
            The path to the save directory.

        """
        if isinstance(self.experiment, DummyExperiment):
            return None
        return self.experiment.dir


class LightningIRTrainer(Trainer):
    # TODO check that correct callbacks are registered for each subcommand

    def index(
        self,
        model: LightningModule | None = None,
        dataloaders: Any | LightningDataModule | None = None,
        ckpt_path: str | Path | None = None,
        verbose: bool = True,
        datamodule: LightningDataModule | None = None,
    ) -> List[Mapping[str, float]]:
        """Index a collection of documents."""
        return super().test(model, dataloaders, ckpt_path, verbose, datamodule)

    def search(
        self,
        model: LightningModule | None = None,
        dataloaders: Any | LightningDataModule | None = None,
        ckpt_path: str | Path | None = None,
        verbose: bool = True,
        datamodule: LightningDataModule | None = None,
    ) -> List[Mapping[str, float]]:
        """Search for relevant documents."""
        return super().test(model, dataloaders, ckpt_path, verbose, datamodule)

    def re_rank(
        self,
        model: LightningModule | None = None,
        dataloaders: Any | LightningDataModule | None = None,
        ckpt_path: str | Path | None = None,
        verbose: bool = True,
        datamodule: LightningDataModule | None = None,
    ) -> List[Mapping[str, float]]:
        """Re-rank a set of retrieved documents."""
        return super().test(model, dataloaders, ckpt_path, verbose, datamodule)


class LightningIRCLI(LightningCLI):
    @staticmethod
    def configure_optimizers(
        lightning_module: LightningModule,
        optimizer: torch.optim.Optimizer,
        lr_scheduler: WarmupLRScheduler | None = None,
    ) -> Any:
        if lr_scheduler is None:
            return optimizer

        return [optimizer], [{"scheduler": lr_scheduler, "interval": lr_scheduler.interval}]

    def add_arguments_to_parser(self, parser):
        parser.add_lr_scheduler_args(tuple(LR_SCHEDULERS))
        parser.link_arguments("model.init_args.model_name_or_path", "data.init_args.model_name_or_path")
        parser.link_arguments("model.init_args.config", "data.init_args.config")
        parser.link_arguments("model", "data.init_args.model", apply_on="instantiate")
        parser.link_arguments("trainer.max_steps", "lr_scheduler.init_args.num_training_steps")

    @staticmethod
    def subcommands() -> Dict[str, Set[str]]:
        return {
            "fit": LightningCLI.subcommands()["fit"],
            "index": {"model", "dataloaders", "datamodule"},
            "search": {"model", "dataloaders", "datamodule"},
            "re_rank": {"model", "dataloaders", "datamodule"},
        }


def main():
    """
    generate config using `python main.py fit --print_config > config.yaml`
    additional callbacks at:
    https://lightning.ai/docs/pytorch/stable/api_references.html#callbacks

    Example:
        To obtain a default config:

            python main.py fit \
                --trainer.callbacks=ModelCheckpoint \
                --optimizer AdamW \
                --trainer.logger LightningIRWandbLogger \
                --print_config > default.yaml

        To run with the default config:

            python main.py fit \
                --config default.yaml

    """
    LightningIRCLI(
        trainer_class=LightningIRTrainer,
        save_config_callback=LightningIRSaveConfigCallback,
        save_config_kwargs={"config_filename": "pl_config.yaml", "overwrite": True},
    )


if __name__ == "__main__":
    main()



from typing import Dict, Literal, Sequence, Type

from transformers import BatchEncoding

from ...cross_encoder.tokenizer import CrossEncoderTokenizer
from .config import T5CrossEncoderConfig


class T5CrossEncoderTokenizer(CrossEncoderTokenizer):

    config_class: Type[T5CrossEncoderConfig] = T5CrossEncoderConfig

    def __init__(
        self,
        *args,
        query_length: int = 32,
        doc_length: int = 512,
        decoder_strategy: Literal["mono", "rank"] = "mono",
        **kwargs,
    ):
        super().__init__(
            *args, query_length=query_length, doc_length=doc_length, decoder_strategy=decoder_strategy, **kwargs
        )
        self.decoder_strategy = decoder_strategy

    def tokenize(
        self,
        queries: str | Sequence[str] | None = None,
        docs: str | Sequence[str] | None = None,
        num_docs: Sequence[int] | None = None,
        **kwargs,
    ) -> Dict[str, BatchEncoding]:
        expanded_queries, docs = self.preprocess(queries, docs, num_docs)
        if self.decoder_strategy == "mono":
            pattern = "Query: {query} Document: {doc} Relevant:"
        elif self.decoder_strategy == "rank":
            pattern = "Query: {query} Document: {doc}"
        else:
            raise ValueError(f"Unknown decoder strategy: {self.decoder_strategy}")
        input_texts = [pattern.format(query=query, doc=doc) for query, doc in zip(expanded_queries, docs)]

        return_tensors = kwargs.get("return_tensors", None)
        if return_tensors is not None:
            kwargs["pad_to_multiple_of"] = 8
        return {"encoding": self(input_texts, **kwargs)}



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, List, Sequence, Tuple, Type

import torch

from ..bi_encoder.model import BiEncoderEmbedding

if TYPE_CHECKING:
    from ..bi_encoder import BiEncoderModule, BiEncoderOutput


class Searcher(ABC):
    def __init__(
        self, index_dir: Path | str, search_config: SearchConfig, module: BiEncoderModule, use_gpu: bool = True
    ) -> None:
        super().__init__()
        self.index_dir = Path(index_dir)
        self.search_config = search_config
        self.module = module
        self.device = torch.device("cuda") if use_gpu and torch.cuda.is_available() else torch.device("cpu")

        self.doc_ids = (self.index_dir / "doc_ids.txt").read_text().split()
        self.doc_lengths = torch.load(self.index_dir / "doc_lengths.pt")

        self.to_gpu()

        self.num_docs = len(self.doc_ids)
        self.cumulative_doc_lengths = torch.cumsum(self.doc_lengths, dim=0)

        if self.doc_lengths.shape[0] != self.num_docs or self.doc_lengths.sum() != self.num_embeddings:
            raise ValueError("doc_lengths do not match index")

    def to_gpu(self) -> None:
        self.doc_lengths = self.doc_lengths.to(self.device)

    @property
    @abstractmethod
    def num_embeddings(self) -> int: ...

    @abstractmethod
    def _search(self, query_embeddings: BiEncoderEmbedding) -> Tuple[torch.Tensor, torch.Tensor, List[int]]: ...

    def _filter_and_sort(
        self,
        doc_scores: torch.Tensor,
        doc_idcs: torch.Tensor | None,
        num_docs: Sequence[int] | None,
    ) -> Tuple[torch.Tensor, List[str], List[int]]:
        if (doc_idcs is None) != (num_docs is None):
            raise ValueError("doc_ids and num_docs must be both None or not None")
        if doc_idcs is None and num_docs is None:
            # assume we have searched the whole index
            k = min(self.search_config.k, doc_scores.shape[0])
            values, idcs = torch.topk(doc_scores.view(-1, self.num_docs), k)
            num_queries = values.shape[0]
            values = values.view(-1)
            idcs = idcs.view(-1)
            doc_ids = [self.doc_ids[doc_idx] for doc_idx in idcs.cpu()]
            return values, doc_ids, [k] * num_queries

        assert doc_idcs is not None and num_docs is not None
        per_query_doc_scores = torch.split(doc_scores, num_docs)
        per_query_doc_idcs = torch.split(doc_idcs, num_docs)
        new_num_docs = []
        _doc_scores = []
        doc_ids = []
        for query_idx, scores in enumerate(per_query_doc_scores):
            k = min(self.search_config.k, scores.shape[0])
            values, idcs = torch.topk(scores, k)
            _doc_scores.append(values)
            doc_ids.extend([self.doc_ids[doc_idx] for doc_idx in per_query_doc_idcs[query_idx][idcs].cpu()])
            new_num_docs.append(k)
        doc_scores = torch.cat(_doc_scores)
        return doc_scores, doc_ids, new_num_docs

    def search(self, output: BiEncoderOutput) -> Tuple[torch.Tensor, List[str], List[int]]:
        query_embeddings = output.query_embeddings
        if query_embeddings is None:
            raise ValueError("Expected query_embeddings in BiEncoderOutput")
        doc_scores, doc_idcs, num_docs = self._search(query_embeddings)
        doc_scores, doc_ids, num_docs = self._filter_and_sort(doc_scores, doc_idcs, num_docs)

        return doc_scores, doc_ids, num_docs


class SearchConfig:
    search_class: Type[Searcher] = Searcher

    def __init__(self, k: int = 10) -> None:
        self.k = k



from pathlib import Path
from typing import Sequence

import pytest
from transformers import AutoModel

from lightning_ir.base import LightningIRModel, LightningIRModule
from lightning_ir.cross_encoder import CrossEncoderModule
from lightning_ir.data import LightningIRDataModule, RunDataset, TupleDataset
from lightning_ir.loss.loss import InBatchLossFunction
from lightning_ir.main import LightningIRTrainer

DATA_DIR = Path(__file__).parent / "data"


def tuples_datamodule(module: LightningIRModule, inference_datasets: Sequence[RunDataset]) -> LightningIRDataModule:
    datamodule = LightningIRDataModule(
        module=module,
        num_workers=0,
        train_batch_size=3,
        inference_batch_size=2,
        train_dataset=TupleDataset("lightning-ir", targets="order", num_docs=2),
        inference_datasets=inference_datasets,
    )
    datamodule.setup(stage="fit")
    return datamodule


def test_training_step(train_module: LightningIRModule, inference_datasets: Sequence[RunDataset]):
    if train_module.loss_functions is None:
        pytest.skip("Loss function is not set")
    datamodule = tuples_datamodule(train_module, inference_datasets)
    dataloader = datamodule.train_dataloader()
    batch = next(iter(dataloader))
    if isinstance(train_module, CrossEncoderModule) and any(
        isinstance(loss_function[0], InBatchLossFunction) for loss_function in train_module.loss_functions
    ):
        with pytest.raises(RuntimeError):
            loss = train_module.training_step(batch, 0)
        return
    loss = train_module.training_step(batch, 0)
    assert loss


def test_validation(train_module: LightningIRModule, inference_datasets: Sequence[RunDataset]):
    datamodule = tuples_datamodule(train_module, inference_datasets)
    dataloader = datamodule.val_dataloader()[0]
    for batch, batch_idx in zip(dataloader, range(2)):
        train_module.validation_step(batch, batch_idx, 0)

    trainer = LightningIRTrainer(logger=False)
    trainer.validate(model=train_module, dataloaders=dataloader)

    metrics = trainer.callback_metrics
    assert metrics is not None
    for key, value in metrics.items():
        if not key:
            continue
        metric = key.split("/")[1]
        assert metric in {"nDCG@10"} or "validation" in metric
        assert value


def test_seralize_deserialize(module: LightningIRModule, tmp_path: Path):
    model = module.model
    save_dir = str(tmp_path / model.config_class.model_type)
    model.save_pretrained(save_dir)
    new_models = [
        model.__class__.from_pretrained(save_dir),
        model.__class__.__bases__[0].from_pretrained(save_dir),
        LightningIRModel.from_pretrained(save_dir),
        AutoModel.from_pretrained(save_dir),
    ]
    for new_model in new_models:
        for key, value in model.config.__dict__.items():
            if key in (
                "torch_dtype",
                "_name_or_path",
                "_commit_hash",
                "transformers_version",
                "model_type",
            ):
                continue
            assert getattr(new_model.config, key) == value
        for key, value in model.state_dict().items():
            assert new_model.state_dict()[key].equal(value)



import array
from pathlib import Path

import torch

from ..bi_encoder import BiEncoderConfig, BiEncoderOutput
from ..data import IndexBatch
from .indexer import IndexConfig, Indexer


class SparseIndexer(Indexer):
    def __init__(
        self,
        index_dir: Path,
        index_config: "SparseIndexConfig",
        bi_encoder_config: BiEncoderConfig,
        verbose: bool = False,
    ) -> None:
        super().__init__(index_dir, index_config, bi_encoder_config, verbose)
        self.crow_indices = array.array("L")
        self.crow_indices.append(0)
        self.col_idcs = array.array("I")
        self.values = array.array("f")

    def add(self, index_batch: IndexBatch, output: BiEncoderOutput) -> None:
        doc_embeddings = output.doc_embeddings
        if doc_embeddings is None:
            raise ValueError("Expected doc_embeddings in BiEncoderOutput")

        doc_lengths = doc_embeddings.scoring_mask.sum(dim=1)
        embeddings = doc_embeddings.embeddings[doc_embeddings.scoring_mask]
        num_docs = len(index_batch.doc_ids)
        self.doc_ids.extend(index_batch.doc_ids)

        token_idcs, dim_idcs = torch.nonzero(embeddings, as_tuple=True)
        crow_indices = token_idcs.bincount().cumsum(0) + self.crow_indices[-1]
        values = embeddings[token_idcs, dim_idcs]
        self.crow_indices.extend(crow_indices.cpu().tolist())
        self.col_idcs.extend(dim_idcs.cpu().tolist())
        self.values.extend(values.cpu().tolist())

        self.doc_lengths.extend(doc_lengths.cpu().tolist())
        self.num_embeddings += embeddings.shape[0]
        self.num_docs += num_docs

    def to_gpu(self) -> None:
        pass

    def to_cpu(self) -> None:
        pass

    def save(self) -> None:
        super().save()
        index = torch.sparse_csr_tensor(
            torch.tensor(self.crow_indices),
            torch.tensor(self.col_idcs),
            torch.tensor(self.values),
            torch.Size([self.num_embeddings, self.bi_encoder_config.embedding_dim]),
        )
        torch.save(index, self.index_dir / "index.pt")


class SparseIndexConfig(IndexConfig):
    indexer_class = SparseIndexer



from typing import Tuple

import torch
from transformers.models.bert.modeling_bert import BertSelfAttention

try:
    from flash_attn import flash_attn_func
except ImportError:
    flash_attn_func = None


def flash_attention_forward(
    self: BertSelfAttention,
    hidden_states: torch.Tensor,
    attention_mask: torch.FloatTensor | None,
    *args,
    **kwargs,
) -> Tuple[torch.Tensor]:
    query = self.transpose_for_scores(self.query(hidden_states))
    key = self.transpose_for_scores(self.key(hidden_states))
    value = self.transpose_for_scores(self.value(hidden_states))

    if attention_mask is not None and not attention_mask.any():
        attention_mask = None

    if flash_attn_func is not None and hidden_states.is_cuda and attention_mask is None:
        context = (
            flash_attn_func(
                query.bfloat16().transpose(1, 2),
                key.bfloat16().transpose(1, 2),
                value.bfloat16().transpose(1, 2),
                self.dropout.p if self.training else 0,
            )
            .transpose(1, 2)
            .to(query.dtype)
        )
    else:
        context = torch.nn.functional.scaled_dot_product_attention(
            query,
            key,
            value,
            attention_mask.to(query.dtype) if attention_mask is not None else None,
            self.dropout.p if self.training else 0,
        )

    context = context.permute(0, 2, 1, 3).contiguous()
    new_context_shape = context.size()[:-2] + (self.all_head_size,)
    context = context.view(new_context_shape)
    return (context,)


SELF_ATTENTION_PATTERN = "self"
